# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import json
import os
import configparser
import getpass
import logging as LOG
from powervc_validation_toolkit.lib.utils import get_current_datetime
from powervc_validation_toolkit.services import computeclient, storageclient


class ReadinessJSONParser:
    def __init__(self, file_path):
        self.file_path = file_path

    def setup_logging(self, log_file):
        LOG.basicConfig(
            level=LOG.INFO,
            format="%(asctime)s [%(levelname)s] %(message)s",
            handlers=[
                LOG.FileHandler(log_file),
                # LOG.StreamHandler(sys.stdout)
            ],
        )

    def init_clients_and_logging(self, config_file_path):
        log_file = None
        try:
            current_datetime = get_current_datetime()
            log_file_name = (
                "powervc_readiness" +
                "_" +
                "clean_" +
                str(current_datetime) +
                ".log")
            config = configparser.ConfigParser()
            config_file = config_file_path
            config.read(config_file)

            if "PVC_PASSWORD" in os.environ:
                password = os.environ['PVC_PASSWORD']
            else:
                config_password = config.get('Default', 'pvc_password')
                if config_password.strip() != "":
                    password = config_password
                else:
                    password = getpass.getpass(
                        prompt="Please enter PowerVC password : ")
            self.compute_client = computeclient.ComputeClient(config, password)
            self.storage_client = storageclient.StorageClient(config, password)
            log_location = config.get("settings", "log_location")
            if not os.path.exists(log_location):
                os.makedirs(log_location)
            log_file = os.path.join(log_location, log_file_name)
            self.setup_logging(log_file)
            LOG.info("Started execution of powervc_readiness "
                     " stale resource cleanup")
        except Exception as exc:
            LOG.error(
                f"Error in init_clients_and_logging"
                f" for clean stale resources: {exc}")
        finally:
            return log_file

    def clean_vms(self, list_of_vms):
        for vm in list_of_vms:
            vm_detail_obj = self.compute_client.get_resource("vm", vm)
            if vm_detail_obj:
                LOG.info("VM details derived successfully for cleanup")
                self.compute_client.delete_server(
                    vm_detail_obj)
                is_deleted, err_msg = self.compute_client.wait_for_delete(
                    vm_detail_obj, interval=10)
                if is_deleted:
                    LOG.info(f"VM {vm_detail_obj.id} deleted successfully")
                else:
                    LOG.error(
                        f"Error {err_msg} occurred while"
                        f" deleting VM {vm_detail_obj.id}"
                    )
                    LOG.info(
                        f"Reseting VM to re-attempt delete of"
                        f" {vm_detail_obj.id}"
                    )
                    try:
                        self.compute_client.conn.compute.\
                            reset_server_status(
                                server=vm_detail_obj,
                                state="active"
                            )
                        LOG.info(f"Server {vm_detail_obj.id} is reset "
                                 f" to active state successfully")
                        LOG.info(
                            f"Re-attempting deletion of VM {vm_detail_obj}")
                        self.compute_client.delete_server(vm_detail_obj)
                        is_deleted, err_msg = \
                            self.compute_client.wait_for_delete(
                                vm_detail_obj, interval=10)
                        if is_deleted:
                            LOG.info(
                                f"VM {vm_detail_obj.id} deleted successfully")
                        else:
                            LOG.error(f"Error {err_msg} occurred while "
                                      f"re-attempting deletion of "
                                      f"VM {vm_detail_obj.id}")
                    except Exception as exc:
                        LOG.error(f"Exception occurred in reset and re-attempt"
                                  f" of deletion of VM"
                                  f" {vm_detail_obj.id}:{exc}")
            else:
                LOG.error("Error occurred while deriving VM details")

    def clean_volumes(self, list_of_volumes):
        for volume in list_of_volumes:
            volume_detail_obj = self.compute_client.get_resource(
                "volume", volume)
            if volume_detail_obj:
                LOG.info(
                    f"Volume {volume_detail_obj.id} details derived"
                    f" successfully for cleanup"
                )
                vol_status = volume_detail_obj.status
                if vol_status == "available":
                    LOG.info(
                        f"Volume {volume_detail_obj.id} is in available"
                        f" state and will be deleted"
                    )
                    is_deleted, err_msg = self.storage_client.delete_volume(
                        volume_detail_obj)
                    if is_deleted:
                        LOG.info(f"Volume {volume_detail_obj.id} is "
                                 f" deleted successfully")
                else:
                    LOG.info(f"Volume {volume_detail_obj.id} is not in "
                             f" available state trying to reset the state")
                    try:
                        self.storage_client.conn.block_storage.\
                            reset_volume_status(
                                volume=volume_detail_obj,
                                status="available",
                                attach_status="detached",
                                migration_status=None,
                            )
                        LOG.info(f"Volume {volume_detail_obj.id} is reset "
                                 f" to available state successfully")
                    except Exception as exc:
                        LOG.error(f"Exception occurred in reset "
                                  f" state of volume"
                                  f" {volume_detail_obj.id}:{exc}")

                    # Try to force delete irrespective of reset state is
                    # successful or not.
                    is_deleted, err_msg = self.storage_client.delete_volume(
                        volume_detail_obj)
                    if is_deleted:
                        LOG.info(f"Volume {volume_detail_obj.id} is "
                                 f" deleted successfully")
            else:
                LOG.error(f"Exception occurred while "
                          f" deriving volume {volume} details")

    def parse_json_for_init_client_and_logging(self):
        log_file = None
        try:
            with open(self.file_path, "r") as file:
                self.parsed_data = json.load(file)
                execution_info = self.parsed_data.get("execution_info", {})
                config = execution_info.get("config_file")
                log_file = self.init_clients_and_logging(config)
        except FileNotFoundError:
            LOG.error("File not found.")
            raise Exception
        except Exception as exc:
            LOG.error(
                f"Exception found in parse_json_for_init_client_and_logging: {exc}")
            raise exc
        finally:
            return log_file

    def parse_json(self):
        vm_ids = []
        volume_ids = []
        try:
            test_flow_results = self.parsed_data.get("test_flow_results", {})
            for test_flow in test_flow_results:
                details = test_flow_results.get(test_flow, {})
                flow_details = details.get("details", {})
                for flow in flow_details:
                    tests = flow_details.get(flow, [])
                    if flow.startswith("deploy"):
                        for test in tests:
                            clean_up = test.get('clean_up')
                            if not clean_up or clean_up.get(
                                    'status') != 'success':
                                if test.get("vm_id"):
                                    vm_ids.append(test.get("vm_id"))
                    elif flow.startswith("attach_volume"):
                        for test in tests:
                            clean_up = test.get('clean_up')
                            if not clean_up or clean_up.get(
                                    'status') != 'success':
                                if test.get('volume_id'):
                                    volume_ids.append(test.get("volume_id"))
            LOG.info(f"List of VMs to be deleted: {vm_ids}")
            LOG.info(f"List of volumes to be deleted: {volume_ids}")
            return vm_ids, volume_ids
        except FileNotFoundError:
            LOG.error("File not found.")
            raise Exception
        except json.JSONDecodeError:
            LOG.error("Error decoding JSON.")
            raise Exception
        except Exception as exc:
            LOG.error(f"Exception found in parse_json: {exc}")
            raise exc
