# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import json
import os
import configparser
import getpass
import logging as LOG
from powervc_validation_toolkit.lib.utils import get_current_datetime
from powervc_validation_toolkit.services import computeclient, storageclient


class Precheck_Cleanup:
    def __init__(self, config, compute_client, storage_client):
        self.config = config
        self.compute_client = compute_client
        self.storage_client = storage_client

    def clean_vms(self, list_of_vms):
        try:
            for vm in list_of_vms:
                vm_detail_obj = self.compute_client.get_resource("vm", vm)
                if vm_detail_obj:
                    LOG.info("VM details derived successfully for cleanup")
                    self.compute_client.delete_server(
                        vm_detail_obj)
                    is_deleted, err_msg = self.compute_client.wait_for_delete(
                        vm_detail_obj, interval=10)
                    if is_deleted:
                        LOG.info(f"VM {vm_detail_obj.id} deleted successfully")
                    else:
                        LOG.error(
                            f"Error {err_msg} occurred while"
                            f" deleting VM {vm_detail_obj.id}"
                        )
                        LOG.info(
                            f"Reseting VM to re-attempt delete of"
                            f" {vm_detail_obj.id}"
                        )
                        try:
                            self.compute_client.conn.compute.\
                                reset_server_status(
                                    server=vm_detail_obj,
                                    state="active"
                                )
                            LOG.info(f"Server {vm_detail_obj.id} is reset "
                                     f" to active state successfully")
                            LOG.info(
                                f"Re-attempting deletion of VM {vm_detail_obj}")
                            self.compute_client.delete_server(vm_detail_obj)
                            is_deleted, err_msg = \
                                self.compute_client.wait_for_delete(
                                    vm_detail_obj, interval=10)
                            if is_deleted:
                                LOG.info(
                                    f"VM {vm_detail_obj.id} deleted successfully")
                            else:
                                LOG.error(f"Error {err_msg} occurred while "
                                          f"re-attempting deletion of "
                                          f"VM {vm_detail_obj.id}")
                        except Exception as exc:
                            LOG.error(
                                f"Exception occurred in reset and re-attempt"
                                f" of deletion of VM"
                                f" {vm_detail_obj.id}:{exc}")
                else:
                    LOG.error("Error occurred while deriving VM details")
        except Exception as e:
            LOG.error(
                f"Exception occurred in clean_vms {exc}")

    def clean_volumes(self, list_of_volumes):
        VOL_AVAILABLE_STATUS = "available"
        try:
            for volume_id in list_of_volumes:
                volume_detail_obj = self.compute_client.get_resource(
                    "volume", volume_id)
                if volume_detail_obj:
                    LOG.info(
                        f"Volume {volume_detail_obj.id} details derived"
                        f" successfully for cleanup"
                    )
                    vol_status = volume_detail_obj.status
                    if vol_status == VOL_AVAILABLE_STATUS:
                        LOG.info(
                            f"Volume {volume_detail_obj.id} is in {VOL_AVAILABLE_STATUS}"
                            f" state and will be deleted")
                        is_deleted, err_msg = self.storage_client.delete_volume(
                            volume_detail_obj)
                        if is_deleted:
                            LOG.info(f"Volume {volume_detail_obj.id} is"
                                     f" deleted successfully")
                    else:
                        LOG.info(f"Volume {volume_detail_obj.id} is not in "
                                 f" available state trying to reset the state")
                        try:
                            self.storage_client.conn.block_storage.\
                                reset_volume_status(
                                    volume=volume_detail_obj,
                                    status="available",
                                    attach_status="detached",
                                    migration_status=None,
                                )
                            LOG.info(f"Volume {volume_detail_obj.id} is reset "
                                     f" to available state successfully")
                        except Exception as exc:
                            LOG.error(f"Exception occurred in reset "
                                      f" state of volume"
                                      f" {volume_detail_obj.id}:{exc}")

                        # Try to force delete irrespective of reset state is
                        # successful or not.
                        is_deleted, err_msg = self.storage_client.delete_volume(
                            volume_detail_obj)
                        if is_deleted:
                            LOG.info(f"Volume {volume_detail_obj.id} is "
                                     f" deleted successfully")
                else:
                    LOG.error(f"Exception occurred while "
                              f" deriving volume {volume_id} details")
        except Exception as e:
            LOG.error(
                f"Exception occurred in clean_volumes {exc}")

    def retrieve_stale_vms_from_api(self):
        vm_ids = []
        try:
            default_prefix = self.config.get('Default', 'default_prefix')
            LOG.info(f"Retrieving stale vm with prefix : '{default_prefix}'")
            servers = self.compute_client.conn.compute.servers()
            for server in servers:
                if server.name.startswith(default_prefix):
                    vm_ids.append(server.id)
            if vm_ids:
                LOG.info(
                    f"Triggering cleanup for VM id's : {','.join(vm_ids)}")
            return vm_ids
        except Exception as exc:
            LOG.error(f"Exception found in retrieve_stale_vms_from_api: {exc}")
            raise exc

    def retrieve_stale_volumes_from_api(self):
        volume_ids = []
        try:
            default_prefix = self.config.get('Default', 'default_prefix')
            LOG.info(
                f"Retrieving stale volume with prefix : '{default_prefix}'")
            volumes = self.storage_client.conn.block_storage.volumes()
            for volume in volumes:
                if volume.name.startswith(default_prefix):
                    volume_ids.append(volume.id)
            if volume_ids:
                LOG.info(
                    f"Triggering cleanup for Volume id's : {','.join(volume_ids)}")
            return volume_ids
        except Exception as exc:
            LOG.error(
                f"Exception found in retrieve_stale_volumes_from_api: {exc}")
            raise exc
