# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import argparse
import configparser
import eventlet
# eventlet.monkey_patch()
import getpass
import logging as LOG
import os
import prettytable
import sys
from eventlet import greenthread, GreenPool
from powervc_validation_toolkit.lib.utils import (
    get_current_datetime,
    get_pvc_services_status
)
from powervc_validation_toolkit.services import computeclient, storageclient
from powervc_validation_toolkit.testflows import testflow as test_flow
from powervc_validation_toolkit.results import test_results
from powervc_validation_toolkit.lib import clean_stale_resources
from powervc_validation_toolkit.lib import pre_check_cleanup

MANDATORY_SERVICES = ['virtualip', 'haproxy', 'galera',
                      'rabbitmq-server',
                      'zookeeper', 'httpd',
                      'openstack-glance-api',
                      'openstack-cinder-api',
                      'openstack-cinder-scheduler',
                      'powervc-cinder-conductor', 'neutron-server',
                      'placement-api', 'openstack-swift-proxy',
                      'openstack-swift-account',
                      'openstack-nova-api',
                      'openstack-nova-scheduler',
                      'openstack-nova-conductor',
                      'powervc-cinder-ibm-health',
                      'powervc-neutron-ibm-health',
                      'powervc-nova-ibm-health'
                      ]

ACTIVE_PASSIVE_SERVICES = [
    'ttv-validation-api',
    'panko-api',
    'powervc-bumblebee',
    'powervc-cinder-ibm-health',
    'powervc-nova-ibm-health',
    'powervc-neutron-ibm-health',
    'openstack-ceilometer-notification',
    'openstack-ceilometer-polling',
    'openstack-nova-ibm-ego-ha-service',
    'openstack-nova-ibm-ego-resource-optimization'
]


class PowerVCReadiness:
    ''' Main class for powervc_readiness binary'''

    def __init__(self, config_file):
        """Initialize config parser and loads config"""
        try:
            self.config = configparser.ConfigParser()
            self.config_file = config_file
            self.load_config()

        except Exception as exc:
            LOG.error(f'Unable to start powervc readiness '
                      f'script due to error: {exc}')

    def validate_section(self, test_section):
        """Validates a particular test section for valid values"""
        # LOG.info(f"Validate section : {test_section}")
        return_code = 0
        mandatory_keys = [
            "image_name",
            "network_name",
            "compute_template",
            "host_group"]
        non_mandatory_keys = [
            "use_custom_template_values",
            "custom_compute_template"]
        try:
            missing_keys = []
            keys_with_empty_values = []
            if test_section in self.config.sections():
                test_section_dict = dict(self.config.items(test_section))
                if test_section.startswith("deploy"):
                    for key in mandatory_keys:
                        if key not in test_section_dict:
                            if key == 'host_group':
                                if test_section_dict.get(
                                        "deploy_in_all_hosts_in_pvc", None):
                                    pass
                                else:
                                    missing_keys.append(key)
                            else:
                                missing_keys.append(key)
                    if missing_keys:
                        LOG.error(
                            f"Mandatory key:'{missing_keys}' not "
                            f"found in '{test_section}'"
                        )
                        raise Exception

                    # Additional validation to check if
                    # deploy_in_all_hosts_in_pvc and
                    # deploy_in_all_hosts_in_hg are False and user has not
                    # provided host param in conf. Host is optional param
                    # if any of those flags are set to True. But if both
                    # are false, then host is required.
                    depl_all_host_in_pvc = test_section_dict.get(
                        "deploy_in_all_hosts_in_pvc", "")
                    depl_all_host_in_hg = test_section_dict.get(
                        "deploy_in_all_hosts_in_hg", "")
                    host_name = test_section_dict.get("host", "")
                    if (
                        not any([depl_all_host_in_pvc, depl_all_host_in_hg,
                                 host_name]
                                )
                        or
                        (
                            depl_all_host_in_pvc == "False" and
                            depl_all_host_in_hg == "False" and not host_name
                        )
                    ):
                        err_msg = (
                            f"You have not set deploy_in_all_hosts_in_hg or "
                            f"deploy_in_all_hosts_in_pvc to True. "
                            f"In that case host is mandatory param for "
                            f"deploy test. Please update '{test_section}' "
                            f"with correct values and try again."
                        )
                        LOG.error(err_msg)
                        raise Exception(err_msg)

                for key in test_section_dict:
                    if not test_section_dict.get(
                            key, None) and key not in non_mandatory_keys:
                        keys_with_empty_values.append(key)
                if keys_with_empty_values:
                    LOG.error(
                        f"Config section '{test_section}' has empty "
                        f"values for keys: {keys_with_empty_values}"
                    )
                    raise Exception
                else:
                    LOG.info(
                        f"Successful validation of config section :"
                        f" '{test_section}'")
            else:
                LOG.error(
                    f"Config section named '{test_section}' is not "
                    f"a valid section in the config file."
                )
                raise Exception
        except Exception as e:
            LOG.error(
                f"Exception encountered while validating "
                f"config section '{test_section}'. {e}."
            )
            return_code = 1
        return return_code

    def _check_for_custom_template_section(self, tests_list):
        try:
            for tests in tests_list:
                if tests.startswith("deploy"):
                    if self.config.has_option(tests,
                                              "use_custom_template_values"):
                        custom_compute_template_param_value = self.config.get(
                            tests, "use_custom_template_values"
                        )
                        if custom_compute_template_param_value in [
                                "True", "TRUE", "true"]:
                            if self.config.has_option(
                                    tests, "custom_compute_template"):
                                compute_template_section_name = \
                                    self.config.get(
                                        tests, "custom_compute_template"
                                    )
                                tests_list.append(
                                    compute_template_section_name
                                )
                            else:
                                err_msg = (
                                    f"Key 'custom_compute_template' is a "
                                    f"mandatory parameter if "
                                    f"'use_custom_template_values' is "
                                    f"set to True"
                                )
                                LOG.error(err_msg)
                                raise Exception(err_msg)
            return True
        except Exception as exc:
            err_msg = (
                f"Error found while validating custom "
                f"compute template: {exc}."
            )
            LOG.error(err_msg)
            return False

    def validate_conf(self, testflow):
        LOG.info(f"Started config validation for {testflow}")
        invalid_conf_msg = ("Conf file has validation failures. "
                            "Please check the logs for detailed errors. "
                            "Fix the conf file and you can rerun "
                            "the readiness again.")
        tests = self.config.get('TestFlows', testflow)
        tests_list = tests.replace(" ", "").split(",")
        if not tests_list[0]:
            msg = ("There are no tests mentioned under testflow"
                   " in the conf.")
            LOG.error(msg)
            return False, msg
        exec_tests = [test for n, test in enumerate(tests_list) if test not in
                      tests_list[:n]]
        if len(exec_tests) < len(tests_list):
            LOG.info("There are repeated tests. In this case it will be "
                     "executed only once. ")
        exec_tests.append('Default')
        is_ct_section_valid = self._check_for_custom_template_section(
            exec_tests
        )
        pool = GreenPool()
        return_codes = pool.starmap(
            self.validate_section,
            [(test_name,) for test_name in exec_tests])
        if any([return_code != 0
                for return_code in return_codes]) or not is_ct_section_valid:
            return False, invalid_conf_msg
        else:
            return True, None

    def get_common_tests(self, testflow_list):
        """Derives common test section from all testflows in list"""
        # The tests_list can have duplicate test names
        # which is removed during consumption in validate_conf_all
        tests_list = []
        for testflow in testflow_list:
            tests = self.config.get('TestFlows', testflow)
            if not tests:
                # No tests mentioned in conf
                continue
            # Remove space if provided between test names
            tests_list_from_flow = tests.replace(" ", "").split(",")
            uniq_test = set(tests_list_from_flow)
            if len(tests_list_from_flow) != len(uniq_test):
                LOG.info(f"There are repeated tests in "
                         "testflow {testflow}. In this case the test "
                         "will be executed only once.")
            tests_list.extend(tests_list_from_flow)
        return tests_list

    def validate_conf_all(self, testflow_list):
        """Validate all flows conf file"""
        LOG.info(f"Started config validation for {testflow_list}")
        invalid_conf_msg = ("Conf file has validation failures. "
                            "Please check the logs for detailed errors. "
                            "Fix the conf file and you can rerun "
                            "the readiness again.")
        tests_list = self.get_common_tests(testflow_list)
        # removes duplicate tests in testflow
        exec_tests = [test for n, test in enumerate(tests_list) if test not in
                      tests_list[:n]]
        if not exec_tests:
            msg = ("There are no tests mentioned under testflow"
                   " in the conf.")
            LOG.error(msg)
            return False, msg
        exec_tests.append('Default')
        is_ct_section_valid = self._check_for_custom_template_section(
            exec_tests)
        pool = GreenPool()
        return_codes = pool.starmap(
            self.validate_section,
            [(test_name,) for test_name in exec_tests])
        if any([return_code != 0
                for return_code in return_codes]) or not is_ct_section_valid:
            return False, invalid_conf_msg
        else:
            return True, None

    def get_services_status_from_pvc_env(self):
        """
        gets status output from powervc env. Uses http call
        :return: dict
        """
        pvc_ip = self.config.get("Default", "pvc_ip")
        pvc_srvcs_status_port = "9402"
        pvc_services_json_data = get_pvc_services_status(pvc_ip,
                                                         pvc_srvcs_status_port)
        if pvc_services_json_data is not None:
            return pvc_services_json_data
        return None

    def column_pad(self, *columns):
        """
        method to add empty rows to keep num of rows in all the col intact
        add rows based on col with max number of rows
        :param columns:
        :return: None
        """
        max_len = max([len(c) for c in columns])
        for c in columns:
            c.extend([''] * (max_len - len(c)))

    def display_validation_output(self, services):
        """
        Show the services dict in tabular format
        :param services:
        :param table:
        :return: None
        """
        pt = prettytable.PrettyTable()
        pt.align = "r"
        # tab header
        col = ["Mandatory Services", "Necessary Services", "Other Services"]

        # convert {service:status} to list to add as column
        failed_man = [f"{k} : {v}" for k, v in services['mandatory'].items()]
        failed_nec = [f"{k} : {v}" for k, v in services['necessary'].items()]
        failed_oth = [f"{k} : {v}" for k, v in services['other'].items()]
        # add empty rows to keep all column length equal
        self.column_pad(failed_man, failed_nec, failed_oth)

        # add cols if rows not empty
        # if the first ele of list is empty obvious the list has no data
        if len(failed_man[0]) != 0:
            pt.add_column(col[0], failed_man)
        if len(failed_nec[0]) != 0:
            pt.add_column(col[1], failed_nec)
        if len(failed_oth[0]) != 0:
            pt.add_column(col[2], failed_oth)
        LOG.info(pt)

    def validate_services_status(self):
        """
        Method to find if there are any services with status other than started
        Uses powervc-services status cmd
        If status_data is not none
            1. Check for type of services --> mandatory and others
            2. Multinode:
                a. if mand service failed on all node stop exec
                b. if mand service running atleast on any node cont exe
            3. Singlenode:
                a. check for mand service if running cont else hault
            4. Display service data in tab format
            5. if mand service failed then return True, else False
        If none:
           return None
        :return: Boolean or None
        """
        # TODO: get hardcoded powervc-services status val. Once https is
        #  implemented remove this code and use request url
        status_data = self.get_services_status_from_pvc_env()
        any_mandatory_services_failed_on_all_nodes = []
        if status_data is not None:
            # Actual code starts. Just replace status_data if needed
            service_not_started = {'mandatory': {},
                                   'necessary': {},
                                   'other': {}}
            for host in status_data:
                for service in status_data.get(host):
                    srvc_status = status_data.get(host). \
                        get(service).get('status')
                    if isinstance(srvc_status, list):
                        srvc_status = srvc_status[0]
                    if srvc_status.lower() != 'started':
                        if (
                            service in ACTIVE_PASSIVE_SERVICES
                            and srvc_status.lower() == 'na'
                        ):
                            continue
                        if service in MANDATORY_SERVICES:
                            col_name = 'mandatory'
                        elif (service.startswith('openstack-nova-compute') or
                              service.startswith('openstack-cinder-volume')):
                            if srvc_status.lower() == 'na':
                                continue
                            col_name = 'necessary'
                        else:
                            col_name = 'other'
                        service_not_started[
                            f'{col_name}'].update(
                            {service: status_data[host][service]['status']})
                if len(service_not_started.get('mandatory')) > 0:
                    any_mandatory_services_failed_on_all_nodes.append(
                        service_not_started.get('mandatory'))
                if (
                        len(service_not_started.get('mandatory')) > 0 or
                        len(service_not_started.get('necessary')) > 0 or
                        len(service_not_started.get('other')) > 0
                ):
                    LOG.error(f"Below services are not started on host:"
                              f" {host}")
                    self.display_validation_output(service_not_started)
            # LOG.info(f"mandatory failed services:"
            #          f"{any_mandatory_services_failed_on_all_nodes}")

            # any_mandatory_services_failed_on_all_nodes is a list that gets
            # updated for each mand service failed on a node. Ideally it
            # conts dict of failed mand services. So, if a service is failed
            # on all nodes then the dict len will be same. Here status_data
            # len of len of hosts. Hence comparing
            # any_mandatory_services_failed_on_all_nodes and len of status_data
            return True if len(any_mandatory_services_failed_on_all_nodes) == \
                           len(status_data) else False
        return None

    def init_client_and_results(self, password, json_file):
        """
        To initialize readiness result class and
        compute client, storage client.
        :return: True/False
        """
        try:
            self.readiness_result = test_results.ReadinessResult(
                json_file, self.config_file)
            self.compute_client = computeclient.ComputeClient(
                self.config, password)
            self.storage_client = storageclient.StorageClient(
                self.config, password)
            return True, "success"
        except Exception as exc:
            error_msg = ("Error in initialization of "
                         "openstack client. Exception is " +
                         str(exc))
            LOG.error(
                f'Error in initialization client and readiness result: {exc}')
            return False, error_msg

    def load_config(self):
        self.config.read(self.config_file)

    def run_test_flow(self, testflow):
        try:
            LOG.info(f"Started execution of testflow {testflow}")
            tests = self.config.get('TestFlows', testflow)
            if not tests:
                # No tests mentioned in conf file
                LOG.info(f"The testflow {testflow} has no tests to execute")
                return
            tests_list = tests.replace(" ", "").split(",")
            # removes duplicate tests in testflow
            exec_tests = [test for n, test in enumerate(tests_list) if test
                          not in tests_list[:n]]
            # Test flow can't start with attach_volume
            if exec_tests[0].startswith('attach_volume'):
                LOG.error('Test should start with deploy. attach_volume test '
                          'can be run only after deploy test in a test flow. '
                          'Please correct the conf and re execute the flow')
                return
            # Instantiate TestFlow class
            t_flow = test_flow.TestFlow(
                self.config,
                self.compute_client,
                self.storage_client,
                testflow,
                exec_tests)

            # Execute a test flow
            t_flow.setUp(self.readiness_result)
            LOG.info(f"Running update_flow_results for testflow {testflow}")
            t_flow.update_flow_results(self.readiness_result)
            LOG.info(f"Running tear_down for testflow {testflow}")
            t_flow.tear_down()
            LOG.info(f"Update tear down/clean up results")
            t_flow.update_flow_results(self.readiness_result)
            t_flow.test_flow_result.add_cleanup_summary()
            LOG.info(f"Clean up summary updated")
            t_flow.update_flow_results(self.readiness_result)
            LOG.info(f"Completed execution of testflow {testflow}")
        except Exception as exc:
            LOG.info(f"Execution of testflow {testflow} failed with "
                     f"exception {exc}")

    def validate_flow(self, testflow):
        try:
            self.run_test_flow(testflow)
        except Exception as exc:
            LOG.error(f"Validate flow has failed with exception {exc}")

    def validate_all(self, testflows):
        try:
            green_pool = GreenPool()
            LOG.info("Starting execution of all test flows")
            for flow in testflows:
                green_pool.spawn_n(self.run_test_flow, flow)
            greenthread.sleep(2)
            green_pool.waitall()
            LOG.info("Completed execution all test flows")
        except Exception as exc:
            LOG.error(f"Execution all flows has failed with exception {exc}")


def setup_logging(log_file):
    LOG.basicConfig(
        level=LOG.INFO,
        format="%(asctime)s [%(levelname)s] %(message)s",
        handlers=[
            LOG.FileHandler(log_file),
            # LOG.StreamHandler(sys.stdout)
        ]
    )


def readiness(pvc_readiness, parsed_args, testflows_list):
    try:
        LOG.info("Openstack connection initialized")
        LOG.info(f"Checking PowerVC services status")
        services_not_started = pvc_readiness.validate_services_status()
        if services_not_started is None:
            msg = ("Unable to get PowerVC services status."
                   "Please check the logs for more details.")
            LOG.error(msg)
            LOG.error(f"Readiness tests can not be run. Execution is stopped")
            pvc_readiness.readiness_result.dump_output_file(
                execution_status="failed",
                status_message=msg)
            sys.exit(0)

        elif services_not_started:
            msg = ("PowerVC services which are mandatory for "
                   "readiness test are not running. Please check "
                   "the log file for more details.")
            LOG.error(msg)
            LOG.error(f"Readiness tests can not be run. Execution is stopped")
            pvc_readiness.readiness_result.dump_output_file(
                execution_status="failed",
                status_message=msg)
            sys.exit(0)
        else:
            LOG.info(f"PowerVC services which are mandatory for"
                     f" readiness test are running")
            if parsed_args.run_cleanup_check:
                pre_check_cleanup_handler(
                    pvc_readiness.config,
                    pvc_readiness.compute_client,
                    pvc_readiness.storage_client
                )

        if parsed_args.subcommand == 'testflows':
            if parsed_args.testflow == 'all':
                is_conf_valid, err_msg = pvc_readiness.validate_conf_all(
                    testflows_list)
                if not is_conf_valid:
                    pvc_readiness.readiness_result.dump_output_file(
                        execution_status="failed",
                        status_message=err_msg)
                    sys.exit(1)
                pvc_readiness.validate_all(testflows_list)
                pvc_readiness.readiness_result.dump_output_file(
                    execution_status="completed")
                LOG.info(f"All test flows completed")
            else:
                testflow = parsed_args.testflow
                LOG.info(f"New execution started for testflow - {testflow}")
                if testflow not in testflows_list:
                    invalid_test_flow_msg = ("Invalid test flow name provided")
                    LOG.error(invalid_test_flow_msg)
                    pvc_readiness.readiness_result.dump_output_file(
                        execution_status="failed",
                        status_message=invalid_test_flow_msg)
                else:
                    is_conf_valid, err_msg = pvc_readiness.validate_conf(
                        testflow)
                    if not is_conf_valid:
                        pvc_readiness.readiness_result.dump_output_file(
                            execution_status="failed",
                            status_message=err_msg)
                        sys.exit(1)
                    pvc_readiness.validate_flow(testflow)
                    pvc_readiness.readiness_result.dump_output_file(
                        execution_status="completed")
                    LOG.info(f"Test flow {testflow} completed")

    except Exception as exc:
        LOG.error(f"Exception in readiness is {exc}")


def clean_stale_resources_handler(json_file_path):
    try:
        parse_json_obj = clean_stale_resources.ReadinessJSONParser(
            json_file_path)
        log_file = parse_json_obj.parse_json_for_init_client_and_logging()
        if not log_file:
            LOG.error(f"Error in initialization of client"
                      f" for clean stale resources")
        else:
            try:
                # Fork the process
                pid = os.fork()
                # If pid is zero, it means we're in the child process
                if pid == 0:
                    vm_ids, volume_ids = parse_json_obj.parse_json()
                    if vm_ids:
                        parse_json_obj.clean_vms(vm_ids)
                    else:
                        LOG.info(
                            "No VMs found for cleanup."
                            " Please check if they are already cleaned up.")
                    if volume_ids:
                        parse_json_obj.clean_volumes(volume_ids)
                    else:
                        LOG.info(
                            "No volumes found for cleanup."
                            " Please check if they are already cleaned up.")
                    # print("Child process PID:", os.getpid())
                else:
                    # Parent process
                    # print("Parent process PID:", os.getpid())
                    print("powervc readiness clean execution progress can "
                          " be tracked in log file: ", log_file)
                    # Exit the parent process
                    sys.exit(0)
            except OSError:
                print("Error in forking process. Terminating the execution.")
                sys.exit(1)
    except Exception as exc:
        print("Error in clean stale resources with exc ", exc)


def pre_check_cleanup_handler(config, compute_client, storage_client):
    try:
        pre_check_cleanup_obj = pre_check_cleanup.Precheck_Cleanup(
            config, compute_client, storage_client)
        vm_ids = pre_check_cleanup_obj.retrieve_stale_vms_from_api()
        if vm_ids:
            pre_check_cleanup_obj.clean_vms(vm_ids)
        else:
            LOG.info(
                "No VMs found for cleanup."
                " Please check if they are already cleaned up.")
        volume_ids = pre_check_cleanup_obj.retrieve_stale_volumes_from_api()
        if volume_ids:
            pre_check_cleanup_obj.clean_volumes(volume_ids)
        else:
            LOG.info(
                "No volumes found for cleanup."
                " Please check if they are already cleaned up.")
    except Exception as exc:
        print("pre_check_cleanup_handler with exc ", exc)
        raise exc


def get_version_and_build(file_path):
    try:
        version, build = None, None
        parser = configparser.RawConfigParser()
        parser.read(file_path + '/version.properties')
        if parser.has_section('powervc_readiness'):
            if parser.has_option('powervc_readiness', 'version'):
                version = parser.get('powervc_readiness', 'version')
            if parser.has_option('powervc_readiness', 'build'):
                build = parser.get('powervc_readiness', 'build')
        return version, build
    except Exception as e:
        print("Somthing went wrong while fetching build and version details.")
        raise


def main():
    try:
        current_datetime = get_current_datetime()
        script_dir = os.path.dirname(os.path.abspath(__file__))
        conf_file = os.path.join(script_dir, "conf", "powervc_readiness.conf")

        version, build = get_version_and_build(script_dir)
        formatted_version = ''
        if version and build:
            formatted_version = version + " " + build
        else:
            raise Exception("Unable to fetch version and build details.")

        # Define CLI args
        parser = argparse.ArgumentParser(
            description='PowerVC Readiness CLI')
        parser.add_argument('-V', '--version', action='version',
                            version=formatted_version)
        parser._positionals.title = ('powervc_readiness subcommands')
        parser._optionals.title = ('options')

        subparsers = parser.add_subparsers(
            dest='subcommand', required=True,
            help='Specify the test flow or operation as defined in conf file')

        parser_testflows = subparsers.add_parser(
            'testflows', help=(
                'Execute test flow passed. The param can be '
                'one test flow name defined in conf file or '
                'all keyword which will run all test flows defined '
                'in conf'))
        parser_testflows._positionals.title = ('required')
        parser_testflows._optionals.title = ('options')
        parser_testflows.add_argument(
            '-c', '--conf', help="User specified PowerVC config file path")

        parser_testflows.add_argument(
            '-r', '--run_cleanup_check', help=(
                "To do cleanup of stale resources left behind by any previous "
                "execution before testflow runs."),
            action="store_true",
            default=False)

        parser_testflows.add_argument(
            'testflow', help=(
                "Specify the test flow name as defined in conf file or 'all'. "
                "'all' will execute all test flows defined in conf file."))

        parser_clean = subparsers.add_parser(
            'clean', help=(
                'To do cleanup of stale resources left behind by any previous '
                'execution. You need to pass the readiness result json file '
                'created at that execution as param via --json option '
                'for clean'))
        parser_clean._optionals.title = ('options')
        parser_clean.add_argument('-j', '--json', required=True,
                                  help=("Log Path with file name containing"
                                        " resource detail for cleanup"))

        # Parse the command-line arguments
        parsed_args = parser.parse_args()

        if parsed_args.subcommand == 'clean':
            json_file_path = parsed_args.json
            if not os.path.isfile(json_file_path):
                print("The log file provided for clean option is not"
                      " available. We can not proceed with clean up.")
                sys.exit(1)
            else:
                if json_file_path.endswith(".json"):
                    clean_stale_resources_handler(json_file_path)
                else:
                    print(
                        "The log file provided is invalid extension type. "
                        "It should be of type JSON.")
                    sys.exit(1)
        else:
            conf_file_path = parsed_args.conf
            if (
                not hasattr(parsed_args, 'conf')
                or parsed_args.conf is None
            ):
                # Try to use default conf file
                if not os.path.isfile(conf_file):
                    print(
                        "The conf file for powervc_readiness is not available."
                        " powervc_readiness can not proceed further.")
                    sys.exit(1)
                else:
                    conf_file_path = conf_file
                    pvc_readiness = PowerVCReadiness(conf_file)
            else:
                if not os.path.isfile(conf_file_path):
                    print("The conf file provided for powervc_readiness "
                          " is not available. powervc_readiness can"
                          " not proceed further.")
                    sys.exit(1)
                else:
                    pvc_readiness = PowerVCReadiness(conf_file_path)
            log_file_name = "powervc_readiness" + \
                "_" + str(current_datetime) + ".log"
            json_file_name = "powervc_readiness_results" + \
                "_" + str(current_datetime) + ".json"
            testflows_from_conf = pvc_readiness.config.get(
                'TestFlows', "flows")
            if not testflows_from_conf:
                msg = ("There are no testflows mentioned under flows"
                       " in the conf.")
                LOG.error(msg)
                LOG.error(
                    f"Readiness tests can not be run. Execution is stopped")
                pvc_readiness.readiness_result.dump_output_file(
                    execution_status="failed",
                    status_message=msg)
                sys.exit(0)
            testflows_list = testflows_from_conf.replace(" ", "").split(",")
            log_location = pvc_readiness.config.get('settings', "log_location")

            if not os.path.exists(log_location):
                os.makedirs(log_location)
            log_file = os.path.join(log_location, log_file_name)
            json_file = os.path.join(log_location, json_file_name)
            setup_logging(log_file)

            LOG.info("Started execution of powervc-readiness")
            # pvc_readiness = PowerVCReadiness(conf_file)
            # LOG.info("Openstack connection initialized")

            if "PVC_PASSWORD" in os.environ:
                password = os.environ['PVC_PASSWORD']
            else:
                config_password = pvc_readiness.config.get(
                    'Default', 'pvc_password', fallback="")
                if config_password.strip() != "":
                    password = config_password
                else:
                    password = getpass.getpass(
                        prompt="Please enter PowerVC password : ")
            is_connection_successful, error_msg = \
                pvc_readiness.init_client_and_results(password, json_file)
            if not is_connection_successful:
                conn_err_msg = ("Error in establishing connection to PowerVC."
                                " " + error_msg)
                LOG.error(f"Error in establishing connection to PowerVC : "
                          f"{error_msg}")
                pvc_readiness.readiness_result.dump_output_file(
                    execution_status="failed",
                    status_message=conn_err_msg)
            try:
                # Fork the process
                pid = os.fork()

                # If pid is zero, it means we're in the child process
                if pid == 0:
                    if not is_connection_successful:
                        sys.exit(0)
                    else:
                        pvc_readiness.readiness_result.dump_output_file(
                            execution_status="started")
                        readiness(pvc_readiness, parsed_args, testflows_list)
                    # print("Child process PID:", os.getpid())
                else:
                    # Parent process
                    # print("Parent process PID:", os.getpid())
                    print("powervc readiness execution progress can be tracked"
                          " in log file: ", log_file)
                    print("powervc readiness results can be tracked"
                          " in log file: ", json_file)
                    # Exit the parent process
                    sys.exit(0)
            except OSError:
                print("Error in forking process. Terminating the execution.")
                sys.exit(1)

    except Exception as exc:
        LOG.error(f"Error in main method: {exc}")
        exit(1)


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print("Exception in main invocation", e)
