# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging as LOG
import json


class TestResult:
    def __init__(self, **args):
        for arg in args.keys():
            self.__dict__[arg] = args[arg]


class DeployTestResult(TestResult):
    def __init__(self, test_name=None, status="unknown", message=None,
                 errors=None, vm_name=None, vm_id=None, host_group=None,
                 host_name=None, clean_up={}):
        args = locals()
        del args['self']
        super(DeployTestResult, self).__init__(**args)


class AttachVolumeResult(TestResult):
    def __init__(self, test_name=None, status="unknown", message=None,
                 errors=None, vm_name=None, vm_id=None, volume_name=None,
                 volume_id=None, storage_template_name=None, clean_up={}):
        args = locals()
        del args['self']
        super(AttachVolumeResult, self).__init__(**args)


class MigrateValidateResult(TestResult):
    def __init__(self, test_name=None, status="unknown", message=None,
                 errors=None, vm_name=None, vm_id=None, host=None,
                 target_host=None):
        args = locals()
        del args['self']
        super(MigrateValidateResult, self).__init__(**args)


class TestSummaryResult:
    """
    Test summary result class which includes tests failed
    and completed in a test flow
    """

    def __init__(self, tests_completed=None, tests_failed=None):
        self.tests_completed = {}
        self.tests_in_progress = []
        self.tests_failed = {}
        self.cleanup_status = {}



class TestFlowResult:
    """
    A test flow result class which includes summary
    and details of tests executed.
    """

    def __init__(self, name, summary=None, details=None):
        self.name = name
        self.summary = summary or TestSummaryResult()
        self.details = details or {}

    def update_tests_failed(self, test_name, failure_summary):
        if test_name in self.summary.tests_in_progress:
            self.summary.tests_in_progress.remove(test_name)
        self.summary.tests_failed[test_name] = failure_summary

    def update_tests_in_progress(self, test_name):
        if test_name not in self.summary.tests_in_progress:
            self.summary.tests_in_progress.append(test_name)

    def update_tests_completed(self, test_name, completed_summary):
        if test_name in self.summary.tests_in_progress:
            self.summary.tests_in_progress.remove(test_name)
        self.summary.tests_completed[test_name] = completed_summary

    def add_test_result(self, test_name, result):
        self.details[test_name] = result

    def add_cleanup_summary(self):
        try:
            tes_flow_details = self.details
            for tes in tes_flow_details:
                if not tes.startswith("migrate_validate"):
                    self.summary.cleanup_status[tes]= {}
                    cumulative_cleanup_status_list = []
                    test_details = tes_flow_details.get(tes, [])
                    for details in test_details:
                        cleanup_status = details.get('clean_up')
                        if tes.startswith('deploy'):
                            vm_id = details.get('vm_id')
                            if cleanup_status and vm_id:
                                cumulative_cleanup_status_list.append(cleanup_status.get('status'))
                            elif not cleanup_status and vm_id:
                                cumulative_cleanup_status_list.append('failed')
                        elif tes.startswith("attach_volume"):
                            volume_id = details.get('volume_id')
                            if cleanup_status and volume_id:
                                cumulative_cleanup_status_list.append(cleanup_status.get('status'))
                            elif not cleanup_status and volume_id:
                                cumulative_cleanup_status_list.append('failed')
                    if "failed" in cumulative_cleanup_status_list:
                        self.summary.cleanup_status[tes]['status'] = "failed"
                    else:
                        self.summary.cleanup_status[tes]['status'] = "success"
        except Exception as exc:
            LOG.error(f"Error in adding cleanup summary {exc}")

    def get_flow_result(self):
        res = {}
        res["summary"] = self.summary.__dict__
        res["details"] = self.details
        return res


class ReadinessResult:
    """
    Output json file readiness result class
    flows_results_tuple holds tuple
    (test_flow_name, test_flow_result_object)
    test_flow_result_object - Instance of TestFlowResult
    flows_results : Result dict which will be
    feed to all_results for dumping to output file
    """

    def __init__(self, json_file, config_file):
        self.json_file = json_file
        self.config_file = config_file
        self.flows_results_tuple = {}
        self.flows_results = {}
        self.all_results = {}

    def add_flow_result(self, flow_tuple, result):
        self.flows_results_tuple[flow_tuple] = result

    def get_flow_results_tuple(self):
        return self.flows_results_tuple

    def update_flow_results_for_output(self):
        for (tflow_name, tflow_res) in self.flows_results_tuple:
            self.flows_results[tflow_name] = tflow_res.get_flow_result()

    def dump_output_file(
            self,
            execution_status="in-progress",
            status_message=None):
        try:
            self.update_flow_results_for_output()
            self.all_results["test_flow_results"] = self.flows_results
            execution_info = {}
            execution_info["status"] = execution_status
            execution_info["config_file"] = self.config_file
            if execution_status == "started":
                execution_info["status_message"] = "Program has started."
            if execution_status == "in-progress":
                execution_info["status_message"] = \
                    "Program is currently running."
            elif execution_status == "completed":
                execution_info["status_message"] = \
                    "Program completed successfully."
            else:
                execution_info["status_message"] = status_message

            self.all_results["execution_info"] = execution_info
            try:
                with open(self.json_file, 'w') as f:
                    json.dump(self.all_results, f, indent=4)
            except Exception as exc:
                LOG.error(f"Error occurred while writing "
                          f"results to json file: {exc}")
        except Exception as exc:
            LOG.error(f"Error in updating results to json: {exc}")
