# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from . import openstacksdkclient
import logging as LOG
import requests
import sys
import time
try:
    from openstack import exceptions
except ImportError as import_exc:
    LOG.error(f"Error: Authentication failed with exception: {import_exc}")
    sys.exit('Error: Authentication failed.')
import eventlet


class ComputeClient(openstacksdkclient.OpenStackSdkClient):

    def __init__(self, config, password):
        super().__init__(config, password)
        self.endpoint = self.get_endpoint_url_by_service_name(
            "compute", self.conn.session.get_project_id())

    def host_group_exists(self, host_group):
        """Check whether host group exists in PowerVC"""
        try:
            host_groups = self.conn.compute.aggregates()
            existing_host_group_names = [
                host_group.name for host_group in host_groups]
            return host_group in existing_host_group_names
        except exceptions.OpenStackCloudException as e:
            LOG.info(f"Error checking aggregate existence: {e}")
            return False

    def wait_for_server_active(self, vm, interval=10, wait=900):
        try:
            self.conn.compute.wait_for_server(vm, interval=interval, wait=wait)
            return True, None
        except exceptions.ResourceTimeout as exc:
            LOG.error(
                f"Timedout waiting for VM {vm.name} to come to active state")
            return False, "Timedout waiting for VM to come to active state"
        except exceptions.ResourceFailure as resource_exc:
            LOG.error(f"VM {vm.name} has moved to error state with"
                      f" failure {resource_exc}")
            return False, "VM has moved to error state"
        except Exception as exc:
            LOG.error(f"Waiting for VM {vm.name} to become active has"
                      f" failed with error {exc}")
            return False, "VM has moved to error state"

    def get_hypervisors(self, details=False):
        try:
            hypervisors = self.conn.compute.hypervisors(details=details)
            return list(hypervisors)
        except Exception as exc:
            LOG.error("Unable to fetch hypervisors from compute API. "
                      "Error: %s.", str(exc))
            return []

    def get_aggregates(self):
        try:
            aggs = self.conn.compute.aggregates()
            return True, aggs
        except Exception as exc:
            LOG.error(
                f"Unable to fetch aggregates from compute API. Error is {exc}")
            return False, None

    def find_host_group(self, host_group):
        try:
            hg = self.conn.compute.find_aggregate(host_group)
            return hg
        except Exception as e:
            LOG.error(f"Error in find_host_group: {e}")
            return None

    def list_vms_by_host(self, host):
        try:
            vms = self.conn.compute.servers(host=host)
            return list(vms)
        except Exception as e:
            LOG.error(f"Error in list_vms_by_host: {e}")
            return []

    def find_vm(self, vm_id):
        try:
            vm = self.conn.compute.find_server(vm_id)
            return vm
        except Exception as exc:
            LOG.error(f"Error {exc} in find_vm")
            return None

    def find_host(self, hypervisor_hostname):
        try:
            host = self.conn.compute.find_hypervisor(hypervisor_hostname)
            return host
        except Exception as exc:
            LOG.error(f"Error {exc} in find_host")
            return None

    def spawn_vm(self, name, image_id, flavor, network_id, host):
        """Invoke spawn vm openstack API"""
        try:
            create_server_params = {
                "name": name,
                "image_id": image_id,
                "flavor": flavor,
                "networks": [{"uuid": network_id}],
                "availability_zone": host,
                "timeout": 1800
            }
            server = self.conn.compute.create_server(**create_server_params)
            LOG.info(f"Deploy of VM with ID {server.id} is initiated")
            # self.conn.compute.wait_for_server(server, wait=600)
            return True, "Deploy started successfully", server
        except Exception as e:
            LOG.error(f"Error creating server: {e}")
            err_msg = "Deploy of VM " + name + " on host " + \
                host + " failed with error " + str(e)
            return False, err_msg, None

    def get_vm_faults_by_vm_id(self, vm_id):
        vm_resp = None
        error_details = {}
        try:
            vm_resp = self.pvc_find_vm(vm_id, include_faults=True)
        except Exception as exc:
            for key in ('details', 'message'):
                error_details[key] = str(exc)
        if vm_resp is not None:
            if 'fault' in vm_resp.get('server'):
                fault_detail_lst = vm_resp.get('server').get('fault')
                if fault_detail_lst:
                    fault_det = fault_detail_lst[0]
                    error_details["error_code"] = fault_det.get(
                        'code')
                    error_details["message"] = fault_det.get(
                        'message')
                    fault_details_string = fault_det.get(
                        'details')
                    fault_details = fault_details_string.split("\n")
                    if fault_details and len(fault_details) > 1:
                        error_details["details"] = \
                            fault_details[-2] if fault_details and \
                            fault_details[-2] else fault_details[-3]
                    else:
                        error_details["details"] = fault_details_string
        LOG.info(f"Fault details for VM {vm_id}: {error_details}")
        return error_details

    def pvc_find_vm(self, vm_id, include_faults=False):
        try:
            vm_url = self.endpoint + "/servers" + "/" + vm_id
            if include_faults:
                vm_url = vm_url + "?include_faults"
            response = requests.get(
                vm_url, headers=self.get_headers(), verify=False)
            return response.json()
        except Exception as exc:
            LOG.error(f"Error in pvc_find_vm {exc}")
            raise exc

    def pvc_wait_for_vm(self, vm_id,
                        permissible_statuses=['ACTIVE'],
                        permissible_healths=['OK'],
                        permissible_tasks=[None],
                        interval=10,
                        timeout=1200):
        """
        This method is mostly invoked for vm health to be in OK
        and VM in active state.
        """
        h_errors = {}
        err_msgs = []
        try:
            status = True
            msg = None
            with eventlet.timeout.Timeout(int(timeout)):
                while True:
                    resp = self.pvc_find_vm(vm_id)
                    if resp:
                        vm_status = resp['server']['status']
                        vm_health = \
                            resp['server']['health_status']['health_value']
                        vm_task_state = resp['server']['OS-EXT-STS:task_state']
                        if (
                                vm_status in permissible_statuses and
                                vm_health in permissible_healths and
                                vm_task_state in permissible_tasks):
                            LOG.info(f"VM {vm_id} health is OK")
                            break
                        eventlet.sleep(int(interval))
        except eventlet.timeout.Timeout:
            status = False
            if resp:
                reasons = resp.get('server').get(
                    'health_status').get('value_reason')
                if isinstance(reasons, list):
                    for reason in reasons:
                        res_key = reason.get('resource_property_key')
                        res_val = reason.get('resource_property_value')
                        if res_key is not None and res_val is not None:
                            err_msg = res_key + " is in " + res_val
                        err_msgs.append(err_msg)
                else:
                    err_msgs.append(reasons)
                h_errors["details"] = err_msgs
                msg = (f"VM ({vm_id}) failed to come to desired state "
                       f"with in the required time ({timeout})s. "
                       f"permissible_statuses: {permissible_statuses}, "
                       f"current_vm_status: {vm_status}. "
                       f"permissible_health: {permissible_healths}, "
                       f"current_health : {vm_health}. ")
                LOG.error(msg)
        except Exception as e:
            status = False
            msg = (f"Error in getting health status of VM: {e}")
            LOG.error(msg)
        finally:
            return status, msg, h_errors

    def reset_server(self, server, state):
        try:
            LOG.info(f"Reset of VM {server.id} is invoked")
            self.conn.compute.reset_server_state(server, state)
        except Exception as exc:
            LOG.error(f"Error in reset server for {server.id}"
                      f" Exc is {exc}")

    def delete_server(self, server):
        try:
            LOG.info(f"Deletion of VM {server.id} is invoked")
            self.conn.compute.delete_server(server.id)
        except Exception as exc:
            LOG.error(f"Error in delete server for {server.id}"
                      f" Exc is {exc}")
            raise exc

    def wait_for_delete(self, server, interval, wait=900):
        try:
            LOG.info(f"Waiting for deletion of VM {server.id} to complete.")
            self.conn.compute.wait_for_delete(
                server, interval=interval, wait=wait)
            return True, None
        except exceptions.ResourceTimeout as exc:
            LOG.error(
                f"Timedout waiting for VM {server.name} to get deleted")
            return False, "Timedout waiting for VM to get deleted"

    def pvc_migrate_validate(self, vm_id, host_mtms=None):
        try:
            isSuccess = False
            service_name = "compute"
            endpoint = self.get_endpoint_url_by_service_name(
                service_name, self.conn.session.get_project_id())
            post_data = {"migrate-validate": {"host": host_mtms}}
            url = endpoint + "/servers/" + vm_id + "/action"
            response = requests.post(url, headers=self.get_headers(),
                                     json=post_data, verify=False)
            if response.ok:
                isSuccess = True
                LOG.info("Migrate-Validate has completed successfully")
            else:
                LOG.error("Migrate-Validate has failed")
            return isSuccess, response.json()
        except Exception as exc:
            LOG.error(f"Error in pvc_migrate_validate {exc}")
            msg = str(exc)
            return isSuccess, {'message': msg}
