# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging as LOG
import sys
from novaclient import client as nova_client
try:
    from openstack import connection
    from openstack import exceptions
except ImportError as import_exc:
    LOG.error(f"Error: Authentication failed with exception: {import_exc}")
    sys.exit('Error: Authentication failed.')


class OpenStackSdkClient:
    def __init__(self, config, password):
        self.password = password
        self.auth = self._get_auth_params(config)
        self._init_connection(self.auth)

    def _get_auth_params(self, config):
        try:
            auth_url = config.get('Default', 'auth_url')
            username = config.get('Default', 'pvc_user')
            # password = config.get('Default', 'pvc_password')
            project_name = config.get('Default', 'os_project_name')
            user_domain_name = config.get('Default', 'os_user_domain_name')
            project_domain_name = config.get(
                'Default', 'os_project_domain_name')
            auth = dict(
                auth_url=auth_url,
                username=username,
                password=self.password,
                project_name=project_name,
                user_domain_name=user_domain_name,
                project_domain_name=project_domain_name
            )
        except Exception as exc:
            LOG.error(f"Error: Reading config variables failed"
                      f" with exception: {exc}")
            sys.exit('Error: Authentication Parameter Retrieval Failed.')
        return auth

    def get_headers(self):
        return {
            "X-Auth-Token": self.conn.auth_token,
            "Content-Type": "application/json"}

    def _init_connection(self, auth):
        try:
            # nova_auth = auth
            version = '2.1'
            username = auth.get('username')
            password = auth.get('password')
            auth_url = auth.get('auth_url')
            nova_auth = {
                "insecure": True,
                "project_name": auth.get('project_name'),
                "user_domain_name": auth.get('user_domain_name'),
                "project_domain_name": auth.get('project_domain_name')
            }
            self.conn = connection.Connection(
                auth=auth, verify=False,
                compute_api_version='2.1')
            self.novaclient = nova_client.Client(
                version=version, username=username,
                password=password, auth_url=auth_url,
                **nova_auth)
        except Exception as exc:
            LOG.error(f"Error: Authentication failed with exception: {exc}")
            sys.exit('Error: Authentication failed.')

    def get_resource(self, resource_type, resource_name):
        """Gets a resource from openstack"""
        try:
            if resource_type == 'image':
                resource = self.conn.image.find_image(
                    resource_name, ignore_missing=False)
            elif resource_type == 'flavor':
                resource = self.conn.compute.find_flavor(
                    resource_name, ignore_missing=False)
            elif resource_type == 'network':
                resource = self.conn.network.find_network(
                    resource_name, ignore_missing=False)
            elif resource_type == 'volume':
                resource = self.conn.block_storage.find_volume(
                    resource_name, ignore_missing=False)
            elif resource_type == 'volume_type':
                resource = self.conn.block_storage.find_type(
                    resource_name, ignore_missing=False)
            elif resource_type == 'vm':
                resource = self.conn.compute.find_server(
                    resource_name, ignore_missing=False)
            else:
                LOG.info(f"Error: Invalid resource type '{resource_type}'.")
                return None
            return resource

        except exceptions.ResourceNotFound as e:
            # Handle the exception when the resource is not found
            LOG.error(f"Error: {resource_type.capitalize()} with name"
                      f" {resource_name} not found. {e}")
            return None
        except Exception as e:
            # Handle other exceptions if needed
            LOG.error(f"Error: An unexpected error occurred. {e}")
            return None

    def get_endpoint_url_by_service_name(self, service_name, project_id):
        try:
            all_endpoints = self.conn.identity.endpoints()
            services = self.conn.identity.services()
            service_name_mapping = {}
            for service in services:
                service_name_mapping[service.id] = service.type
            service_id = None
            for svc_id, svc_type in service_name_mapping.items():
                if svc_type == service_name:
                    service_id = svc_id
                    break
            if service_id:
                endpoint = next(
                    (ep for ep in all_endpoints
                        if ep.service_id == service_id),
                    None)
                if endpoint:
                    return endpoint.url.replace("%(tenant_id)s", project_id)
                return None
        except Exception as exc:
            LOG.error(f"Error in get_endpoint_url_by_service_name"
                      " with exception {exc}")
            return None
