# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from . import openstacksdkclient
import logging as LOG
import sys
import requests
try:
    from openstack import exceptions
except ImportError as import_exc:
    LOG.error(f"Error: Authentication failed with exception: {import_exc}")
    sys.exit('Error: Authentication failed.')


class StorageClient(openstacksdkclient.OpenStackSdkClient):
    def __init__(self, config, password):
        super().__init__(config, password)
        self.endpoint = self.get_endpoint_url_by_service_name(
            "volumev3", self.conn.session.get_project_id())

    def create_new_volumes(
            self,
            volume_name,
            volume_size,
            storage_template_id):
        create_volume_request_body = {
            "name": volume_name,
            "size": volume_size,
            "volume_type": storage_template_id
        }
        try:
            vol_details = self.conn.block_storage.create_volume(
                **create_volume_request_body)
            return True, "Volume created", vol_details
        except Exception as exc:
            LOG.error(f"Create volume {volume_name} has failed with"
                      f" exception : {exc}")
            err_msg = ("Creation of volume" + volume_name +
                       " of storage template type " + storage_template_id +
                       "failed with error " + str(exc))
            return False, err_msg, None

    def delete_volume(self, volume, wait=900):
        try:
            LOG.info(f"Deletion of volume {volume.id} is invoked")
            self.conn.block_storage.delete_volume(volume)
            self.conn.block_storage.wait_for_delete(
                volume, interval=10, wait=wait)
            return True, None
        except exceptions.ResourceTimeout as exc:
            LOG.error(
                f"Timedout waiting for volume {volume.id} to get deleted")
            return False, "Timedout waiting for volume to get deleted"
        except Exception as exc:
            LOG.error(f"Error in delete volume for {volume.id}"
                      f" Exc is {exc}")
            err_msg = (f"Error {exc} in deleting volume {volume.id}")
            return False, err_msg

    def wait_for_volume_status(self, volume, status="available",
                               failures=["error"], interval=5, wait=300):
        try:
            self.conn.block_storage.wait_for_status(volume,
                                                    status,
                                                    failures,
                                                    interval,
                                                    wait)
            return True, None
        except exceptions.ResourceTimeout as exc:
            LOG.error(f"Timedout waiting for Volume {volume.name} to come to"
                      f" {status} state")
            return False, ("Timedout waiting for Volume to come to "
                           + status + " state")
        except exceptions.ResourceFailure as resource_exc:
            LOG.error(f"Volume {volume.name} has moved to error state"
                      f" with failure {resource_exc}")
            return False, "Volume has moved to error state"
        except Exception as exc:
            LOG.error(f"Waiting for Volume {volume.name} to become"
                      f" {status} has failed with error {exc}")
            return False, ("Volume has failed to moved to "
                           + status + " state")

    def get_volume_faults_by_volume_id(self, volume_id):
        vol_resp = self.pvc_find_volume_by_id_with_faults(volume_id)
        error_details = {}
        if vol_resp is not None:
            if 'fault' in vol_resp.get('volume'):
                fault_det = vol_resp.get('volume').get('fault')
                error_details["error_code"] = fault_det.get(
                    'code')
                error_details["message"] = fault_det.get(
                    'message')
                error_details["details"] = fault_det.get(
                    'details')
        return error_details

    def pvc_find_volume_by_id_with_faults(self, volume_id):
        try:
            vm_url = self.endpoint + "/volumes" + "/" + \
                volume_id + "?include_faults"
            response = requests.get(
                vm_url, headers=self.get_headers(), verify=False)
            return response.json()
        except Exception as exc:
            LOG.error(f"Error in pvc_find_volume_by_id_with_faults {exc}")
            return None

    """
    def pvc_bulk_volume_create_and_attach(
            self,
            vm_id,
            vol_name,
            vol_size,
            storage_template_type,
            vol_count):
        try:
            service_name = "compute"
            endpoint = self.get_endpoint_url_by_service_name(
                service_name, self.conn.session.get_project_id())
            post_data = {
                "bulkVolumeAttach": {
                    "volumeAttachment": [{
                        "volume": {
                            "name": vol_name,
                            "size": int(vol_size),
                            "volume_type": storage_template_type,
                            "multiattach": False
                        },
                        "count": int(vol_count)
                    }]
                }
            }
            volume_url = endpoint + "/servers" + "/" + vm_id + "/action"
            response = requests.post(
                volume_url,
                headers=self.get_headers(),
                json=post_data,
                verify=False)
            if response.ok:
                LOG.info("Volume attach has been done successfully")
        except Exception as exc:
            LOG.error(f"Error in pvc_bulk_volume_create_and_attach {exc}")
    """
