# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import uuid
import eventlet
# eventlet.monkey_patch()
from eventlet import greenthread, GreenPool
import logging as LOG
from . import base_test


class AttachVolumeTest(base_test.BaseTest):
    def __init__(
            self, name, config, blacklist_hosts,
            compute_client,
            storage_client,
            related_deploy_test,
            test_flow_result,
            readiness_result):
        super(AttachVolumeTest, self).__init__(
            name, config, blacklist_hosts,
            compute_client=compute_client,
            storage_client=storage_client,
            related_deploy_test=related_deploy_test,
            test_flow_result=test_flow_result,
            readiness_result=readiness_result)

    def wait_and_create_and_attach(self, vm):
        try:
            LOG.info(f"Waiting for deploy of VM {vm.name} to complete")
            greenthread.sleep(5)
            build_timeout = self.conf.instance_build_timeout \
                if hasattr(self.conf, 'instance_build_timeout') else 2000
            volume_create_timeout = self.conf.volume_create_timeout \
                if hasattr(self.conf, 'volume_create_timeout') else 2000
            volume_attach_timeout = self.conf.volume_attach_timeout \
                if hasattr(self.conf, 'volume_attach_timeout') else 2000
            is_active, err_msg = self.compute_client.wait_for_server_active(
                vm, wait=int(build_timeout))
            self.related_deploy_test.update_deploy_results(
                vm.id, is_active, err_msg)
            self.related_deploy_test.update_depl_result_to_test_flow_result()
            self.test_flow_result.update_tests_in_progress(self.name)
            self.readiness_result.dump_output_file()
            attach_volum_res_uuid = uuid.uuid4()
            # self.update_deploy_results(vm.id, is_active, err_msg)
            if is_active:
                LOG.info(
                    f"VM {vm.name} became active. Creation of"
                    " new volume and attach is in progress"
                )
                storage_template_type = self.compute_client.get_resource(
                    "volume_type", self.storage_template_name
                )
                if storage_template_type is None:
                    LOG.error(
                        "Error: Volume type not found or unable to fetch "
                        "type details from API"
                    )
                    template_err_msg = (
                        "Error: Volume type not found or unable to fetch "
                        "volume type details from API")
                    vol_result = self.test_results.AttachVolumeResult(
                        test_name=self.name,
                        status="failed",
                        message=template_err_msg,
                        errors={},
                        vm_name=vm.name,
                        vm_id=vm.id,
                        volume_name=None,
                        volume_id=None,
                        storage_template_name=self.storage_template_name)
                    self.results[attach_volum_res_uuid] = vol_result
                    self.update_attach_res_to_test_flow_result()
                    self.readiness_result.dump_output_file()
                    return
                vol_name = None
                if hasattr(self.conf, "default_prefix"):
                    vol_name = self.conf.default_prefix
                if not vol_name:
                    vol_name = self.vol_name + "-" + vm.name
                is_vol_created, msg, vol_detail = \
                    self.storage_client.create_new_volumes(
                        vol_name, self.vol_size, storage_template_type.id)
                if not is_vol_created:
                    faults_details = self.storage_client.\
                        get_volume_faults_by_volume_id(vol_detail.id)
                    errors = {
                        'fault_details': faults_details
                    } if faults_details else {}
                    vol_result = self.test_results.AttachVolumeResult(
                        test_name=self.name,
                        status="failed",
                        message=msg,
                        errors=errors,
                        vm_name=vm.name,
                        vm_id=vm.id,
                        volume_name=vol_name,
                        volume_id=None,
                        storage_template_name=self.storage_template_name
                    )
                    self.results[attach_volum_res_uuid] = vol_result
                    self.update_attach_res_to_test_flow_result()
                    self.readiness_result.dump_output_file()
                else:
                    (
                        is_available,
                        vol_state_msg,
                    ) = self.storage_client.wait_for_volume_status(
                        vol_detail, wait=int(volume_create_timeout))
                    volume_id = vol_detail.id
                    if not is_available:
                        faults_details = self.storage_client.\
                            get_volume_faults_by_volume_id(vol_detail.id)
                        errors = {
                            'fault_details': faults_details
                        } if faults_details else {}
                        vol_result = self.test_results.AttachVolumeResult(
                            test_name=self.name,
                            status="failed",
                            message=vol_state_msg,
                            errors=errors,
                            vm_name=vm.name,
                            vm_id=vm.id,
                            volume_name=vol_name,
                            volume_id=vol_detail.id,
                            storage_template_name=self.storage_template_name
                        )
                        self.results[attach_volum_res_uuid] = vol_result
                        self.update_attach_res_to_test_flow_result()
                        self.readiness_result.dump_output_file()
                        self.tear_down(
                            vm, vol_detail,
                            attach_volum_res_uuid,
                            is_attached=False)
                    else:
                        try:
                            attach_volume_to_server = (
                                self.compute_client.conn
                                .compute.create_volume_attachment(
                                    server=vm, volume_id=volume_id))
                        except Exception as exc:
                            attach_err_msg = (
                                f"Error in create volume attachment for vm {vm.id}"
                                f" Exception is {str(exc)}")
                            LOG.error(attach_err_msg)
                        if attach_volume_to_server is not None:
                            is_attach_successful, err_msg = \
                                self.storage_client.wait_for_volume_status(
                                    vol_detail, status="in-use",
                                    wait=int(volume_attach_timeout))
                            if not is_attach_successful:
                                attach_status = "failed"
                                attach_msg = err_msg
                                faults_details = \
                                    self.storage_client \
                                        .get_volume_faults_by_volume_id(
                                            vol_detail.id)
                                errors = {'fault_details': faults_details}
                            else:
                                attach_status = "success"
                                attach_msg = "Volume attached"
                                errors = {}
                                LOG.info(
                                    f"VM {vm.name} attach of new volume"
                                    " completed successfully")
                            vol_result = self.test_results.AttachVolumeResult(
                                test_name=self.name,
                                status=attach_status,
                                message=attach_msg,
                                errors=errors,
                                vm_name=vm.name,
                                vm_id=vm.id,
                                volume_name=vol_name,
                                volume_id=volume_id,
                                storage_template_name=self.
                                storage_template_name
                            )
                            self.results[attach_volum_res_uuid] = vol_result
                            self.update_attach_res_to_test_flow_result()
                            self.readiness_result.dump_output_file()
                            self.tear_down(
                                vm, vol_detail, attach_volum_res_uuid,
                                is_attached=is_attach_successful)
                        elif attach_err_msg or attach_volume_to_server is None:
                            faults_details = \
                                self.storage_client\
                                    .get_volume_faults_by_volume_id(
                                        vol_detail.id)
                            errors = {
                                'fault_details': faults_details
                            } if faults_details else {}
                            vol_result = self.test_results.AttachVolumeResult(
                                test_name=self.name,
                                status="failed",
                                message=attach_err_msg,
                                errors=errors,
                                vm_name=vm.name,
                                vm_id=vm.id,
                                volume_name=vol_name,
                                volume_id=vol_detail.id,
                                storage_template_name=self.storage_template_name
                            )
                            self.results[attach_volum_res_uuid] = vol_result
                            self.update_attach_res_to_test_flow_result()
                            self.readiness_result.dump_output_file()
                            self.tear_down(
                                vm, vol_detail,
                                attach_volum_res_uuid,
                                is_attached=False)
                            LOG.error("Volume detach skipped since "
                                      " attach to the VM {vm.name} failed")

                # Validate VM health status as well
                LOG.info(f"Waiting for VM {vm.id} health to become OK")
                is_ok, msg, h_errors = self.compute_client.pvc_wait_for_vm(
                    vm.id,
                    permissible_statuses=['ACTIVE'],
                    permissible_healths=['OK'],
                    permissible_tasks=[None],
                    interval=10
                )

                self.related_deploy_test.update_deploy_results(
                    vm.id, is_active, msg,
                    h_errors=h_errors, health_evaluated=True)
                self.related_deploy_test\
                    .update_depl_result_to_test_flow_result()
                self.readiness_result.dump_output_file()

            else:
                vol_result = self.test_results.AttachVolumeResult(
                    test_name=self.name,
                    status="failed",
                    message="Volume attach skipped since VM is not in"
                    " active state",
                    errors={},
                    vm_name=vm.name,
                    vm_id=vm.id,
                    volume_name=None,
                    volume_id=None,
                    storage_template_name=self.storage_template_name)
                self.results[attach_volum_res_uuid] = vol_result
                self.update_attach_res_to_test_flow_result()
                self.readiness_result.dump_output_file()

        except Exception as exc:
            LOG.error(
                f"Exception in wait for server and create and"
                f" attach of volume {exc}"
            )

    def setUp(self):
        try:
            self.storage_template_name = self.conf.storage_template_name
            # self.vol_count = self.conf.volume_count
            self.vol_size = self.conf.volume_size
            vol_name_prefix = self.conf.vol_name_prefix
            self.vol_name = vol_name_prefix
            pool = GreenPool()

            for vm in self.related_deploy_test.get_vms():
                pool.spawn_n(self.wait_and_create_and_attach, vm)
            greenthread.sleep(1)
            pool.waitall()
        except Exception as exc:
            LOG.error(f"Error in setUp for AttachVolumeTest: exc is {exc}")

    def update_attach_res_to_test_flow_result(self):
        attach_result = self.get_results()
        self.test_flow_result.add_test_result(self.name, attach_result)

    def tear_down(
            self,
            vm,
            vol_detail,
            attach_volum_res_uuid,
            is_attached=True):
        # Detach the volume from VM
        try:
            volume_attach_timeout = self.conf.volume_attach_timeout \
                if hasattr(self.conf, 'volume_attach_timeout') else 2000
            delete_timeout = self.conf.delete_timeout \
                if hasattr(self.conf, 'delete_timeout') else 900
            is_detach_successful = False
            if is_attached:
                self.compute_client.conn.compute.delete_volume_attachment(
                    server=vm, volume=vol_detail
                )
                LOG.info(
                    f"VM {vm.name} detach of volume {vol_detail.name}"
                    " completed successfully"
                )
                is_detach_successful, err_msg = \
                    self.storage_client.wait_for_volume_status(
                        vol_detail, wait=int(volume_attach_timeout))
            if is_detach_successful or not is_attached:
                if not is_attached:
                    # Volume is not attached. If volume is stuck, try reset
                    try:
                        self.storage_client.conn.block_storage.\
                            reset_volume_status(
                                volume=vol_detail,
                                status="available",
                                attach_status="detached",
                                migration_status=None,
                            )
                        LOG.info(f"Volume {vol_detail.id} is reset "
                                 f" to available state successfully")
                    except Exception as exc:
                        LOG.error(f"Exception occurred in reset "
                                  f" state of volume"
                                  f" {vol_detail.id}:{exc}")
                is_deleted, del_err_msg = self.storage_client.delete_volume(
                    vol_detail, wait=int(delete_timeout))
                if is_deleted:
                    if is_attached:
                        LOG.info(
                            f"Detached volume '{vol_detail.name}' "
                            f"deleted successfully"
                        )
                        clean_up_msg = ("Volume has been detached and deleted")
                    else:
                        LOG.info(
                            f"Volume '{vol_detail.name}' deleted successfully"
                        )
                        clean_up_msg = ("Volume has been deleted")
                    clean_up_dict = {
                        "status": "success",
                        "message": clean_up_msg
                    }
                else:
                    clean_up_dict = {
                        "status": "failed",
                    }
                    # Volume deletion has failed. Check if any details exist in
                    # volume fault
                    faults_details = \
                        self.storage_client.get_volume_faults_by_volume_id(
                            vol_detail.id)
                    if faults_details:
                        clean_up_dict["errors"] = faults_details
                    else:
                        clean_up_dict["errors"] = {
                            "message": del_err_msg
                        }
                self.results[attach_volum_res_uuid].clean_up = clean_up_dict
            else:
                faults_details = \
                    self.storage_client.get_volume_faults_by_volume_id(
                        vol_detail.id)
                self.results[attach_volum_res_uuid].clean_up = {
                    "status": "failed",
                    "errors": faults_details
                }
                # integrate volume fault and update clean_up
                LOG.error(err_msg)
        except Exception as exc:
            LOG.error(f"Exception in tear_down in detach volume {exc}")

    def get_results(self):
        attach_res_lst = []
        attach_result = {}
        for result in self.results.values():
            attach_result = {
                "status": result.status,
                "message": result.message,
                "volume_name": result.volume_name,
                "volume_id": result.volume_id,
                "vm_name": result.vm_name,
                "vm_id": result.vm_id,
                "storage_template_name": result.storage_template_name,
                "errors": result.errors,
                "clean_up": result.clean_up
            }
            attach_res_lst.append(attach_result)
        return attach_res_lst

    def get_results_summary(self):
        attach_res_lst = []
        attach_result = {}
        attach_len = len(self.results)
        failed_stg_templates = []
        failed_vol_ids = []
        results_summary = {}
        stg_templates = []
        err_vol_cnt = 0
        try:
            if attach_len > 0:
                for result in self.results.values():
                    if result.status == "failed":
                        err_vol_cnt += 1
                        if (
                            result.storage_template_name
                            not in failed_stg_templates
                        ):
                            failed_stg_templates.append(
                                result.storage_template_name)
                        if result.volume_id:
                            # Append volume id only if it exists
                            failed_vol_ids.append(result.volume_id)
                    if result.storage_template_name not in stg_templates:
                        stg_templates.append(result.storage_template_name)
                    attach_result = {
                        "status": result.status,
                        "message": result.message,
                        "volume_name": result.volume_name,
                        "volume_id": result.volume_id,
                        "vm_name": result.vm_name,
                        "vm_id": result.vm_id,
                        "storage_template_name": result.storage_template_name,
                        "errors": result.errors,
                        "clean_up": result.clean_up

                    }
                    attach_res_lst.append(attach_result)
                if err_vol_cnt > 0:
                    msg = (
                        str(err_vol_cnt) +
                        " Out of " +
                        str(attach_len) +
                        " attach volume has failed. Templates where attach failed"
                        " are " +
                        " ".join(failed_stg_templates))
                    results_summary["failed_volume_list"] = failed_vol_ids
                    results_summary["message"] = msg
                    return False, results_summary, attach_res_lst
                else:
                    success_msg = (
                        "Attach of volume with storage template(s) " +
                        " ".join(stg_templates) +
                        " is successful")
                    results_summary["message"] = success_msg
                    return True, results_summary, attach_res_lst
            else:
                return True, results_summary, attach_res_lst
        except Exception as exc:
            LOG.error(f"Error {exc} in getting deploy summary results "
                      f"in attach volume test")
            return True, results_summary, attach_res_lst
