# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging as LOG
import json
from operator import itemgetter
from powervc_validation_toolkit.lib.scg import get_scg_id_by_name
from powervc_validation_toolkit.lib.utils import get_current_datetime
import eventlet
# eventlet.monkey_patch()
import six
import uuid
from eventlet import greenthread, GreenPool
from . import base_test


class DeployTest(base_test.BaseTest):
    def __init__(
            self, name, config, blacklist_hosts,
            compute_client,
            test_flow_result,
            readiness_result):
        super(DeployTest, self).__init__(
            name, config, blacklist_hosts,
            compute_client=compute_client,
            test_flow_result=test_flow_result,
            readiness_result=readiness_result,
            vms=[])
        self.vm_uuid_to_res_uuid_map = {}

    def get_vms(self):
        return self.vms

    def setUp(self):
        try:
            self.image = self.compute_client.get_resource(
                'image', self.conf.image_name)
            if self.image is None:
                LOG.error("Error: Image not found or unable to "
                          "fetch image details from API")
                return False, ("Error: Image not found or unable to "
                               "fetch image details from API")
            block_device_mapping = self.image.properties.get(
                'block_device_mapping')
            block_device_mapping = json.loads(block_device_mapping)
            volume_id = list(
                map(itemgetter('volume_id'), block_device_mapping))
            volume_id = ''.join(map(str, volume_id))
            volume = self.compute_client.get_resource('volume', volume_id)
            if volume is None:
                LOG.error("Error: Image volume not found or unable to "
                          "fetch image volume details from API")
                return False, ("Error: Image volume not found or unable to "
                               "fetch image volume details from API")
            flavor = self.compute_client.get_resource(
                'flavor', self.conf.compute_template)
            if flavor is None:
                LOG.error("Error: Flavor not found or unable to fetch "
                          "flavor details from API")
                return False, ("Error: Flavor not found or unable to fetch "
                               "flavor details from API")
            flavor_vcpus = flavor.vcpus
            flavor_ram = flavor.ram
            self.network = self.compute_client.get_resource(
                'network', self.conf.network_name)
            if self.network is None:
                LOG.error("Error: Network not found or unable to fetch "
                          "flavor details from API")
                return False, ("Error: Flavor not found or unable to fetch "
                               "flavor details from API")

            self.host_group = getattr(self.conf, 'host_group', 'Default Group')
            is_retrieved, aggs = self.compute_client.get_aggregates()
            if not is_retrieved:
                return False, ("Error: Unable to fetch aggregate "
                               "details from API")
            self.aggregates = list(aggs)
            if self.compute_client.host_group_exists(self.host_group):
                LOG.debug(
                    f"The host group '{self.host_group}' exists in PowerVC.")
            else:
                LOG.error(f"The host group '{self.host_group}' does"
                          " not exist in PowerVC. Server creation aborted")
                return False, "The host group does not exist in PowerVC"

            deploy_on_all_hosts_in_hg = \
                self.conf.deploy_in_all_hosts_in_hg if hasattr(
                    self.conf, 'deploy_in_all_hosts_in_hg') else ""
            deploy_in_all_hosts_in_pvc = \
                self.conf.deploy_in_all_hosts_in_pvc if hasattr(
                    self.conf, 'deploy_in_all_hosts_in_pvc') else ""

            if (
                not any([deploy_in_all_hosts_in_pvc,
                         deploy_on_all_hosts_in_hg]
                        )
                    or
                    (deploy_in_all_hosts_in_pvc == "False" and
                     deploy_on_all_hosts_in_hg == "False")
            ):
                # Consider checking host only if
                # deploy_on_all_hosts_in_hg and
                # deploy_in_all_hosts_in_pvc are false.
                self.host = getattr(self.conf, 'host', None)
                if self.host:
                    if not self.compute_client.find_host(self.host):
                        err_msg = (
                            f"Unable to fetch host ({self.host}) details"
                            " from API. Host will not be considered"
                            " for deployment.")
                        LOG.warn(err_msg)
                        # setting it to None
                        self.host = None
                        raise Exception(err_msg)

            use_custom_template = \
                self.conf.use_custom_template_values if hasattr(
                    self.conf, 'use_custom_template_values') else ""
            scg_name = self.conf.storage_connectivity_group
            use_custom_scg = False
            if scg_name != "Any host, all VIOS":
                use_custom_scg = True
            scg_id = get_scg_id_by_name(
                scg_name, self.compute_client.novaclient)
            if scg_id is None:
                LOG.error(f"Error in getting SCG {scg_name}. Please check"
                          " if SCG is valid and nova API service is running")
                return False, ("Error in getting SCG. Please check if SCG is "
                               "valid and nova API service is running")

            if use_custom_template and use_custom_template in [
                    "True", "TRUE", "true"]:
                custom_template_section = \
                    self.conf.custom_compute_template
                custom_template_object = self.get_config_object(
                    custom_template_section)
                # mandatory params
                ram = custom_template_object.mem
                vcpus = custom_template_object.vcpus
                disk = volume.size
                proc_units = custom_template_object.proc_units
                # optional params
                uncapped = getattr(custom_template_object, 'uncapped', "true")
                shared_weight = getattr(
                    custom_template_object, 'shared_weight', "128")
                dedicated_proc = getattr(
                    custom_template_object, 'dedicated_proc', "False")
                ame = getattr(custom_template_object, 'ame', 0)
                srr_capability = getattr(
                    custom_template_object, 'srr_capability', False)
                self.flavor = {
                    "ram": ram,
                    "vcpus": vcpus,
                    "disk": disk,
                    "extra_specs": {
                        "powervm:availability_priority": "127",
                        "powervm:dedicated_proc": dedicated_proc,
                        "powervm:proc_units": proc_units,
                        "powervm:shared_weight": shared_weight,
                        "powervm:uncapped": uncapped,
                        "powervm:storage_connectivity_group": scg_id,
                        "powervm:ame_expansion_factor": ame,
                        "powervm:srr_capability": srr_capability
                    }
                }
            elif use_custom_scg:
                self.flavor = {
                    "ram": flavor_ram,
                    "vcpus": flavor_vcpus,
                    "disk": volume.size,
                    "extra_specs": {
                        "powervm:storage_connectivity_group": scg_id,
                    }
                }
            else:
                self.flavor = flavor

            # Validate if deploy_in_all_hosts_in_pvc is provided
            if (
                deploy_in_all_hosts_in_pvc and
                deploy_in_all_hosts_in_pvc in ["True", "TRUE", "true"]
            ):
                has_hosts = False
                for agg in self.aggregates:
                    hosts = agg.hosts
                    if len(hosts) > 0:
                        has_hosts = True
                        break
                if not has_hosts:
                    return False, ("The host groups does not have any hosts "
                                   "for deploy.")
            elif (
                deploy_on_all_hosts_in_hg
                and deploy_on_all_hosts_in_hg in ["True", "TRUE", "true"]
            ):
                # Validate if the host group provided has hosts for deploy.
                for agg in self.aggregates:
                    if agg.name.lower() == self.host_group.lower():
                        hosts = agg.hosts
                        if len(hosts) == 0:
                            return False, ("The host group provided for "
                                           "deploy does not have any hosts")
            # Setup is done successfully
            return True, "Deploy params validated"
        except Exception as exc:
            msg = (f"Error in setUp for DeployTest: exc is {exc}")
            LOG.error(msg)
            return False, msg

    def spawn_multiple_deploys_and_get_threads(
            self, hosts_lst, host_to_hg_map={}, hg_name=None):
        try:
            gpool = GreenPool()
            threads = {}
            child_threads = []
            vm_name = None
            if hasattr(self.conf, "default_prefix"):
                vm_name = self.conf.default_prefix
            if not vm_name:
                vm_name = self.conf.vm_name_prefix
            deploy_by_hg = False
            if hg_name is not None:
                deploy_by_hg = True
            try:
                for i in range(len(hosts_lst)):
                    if hosts_lst[i] in self.blacklist_hosts:
                        LOG.info(
                            "Host (%s) is in blacklist hosts. So skipping "
                            "deploy on this host." %
                            hosts_lst[i])
                        continue
                    vm_nam = vm_name + "_" + \
                        hosts_lst[i] + "_" + get_current_datetime()
                    child_thread = gpool.spawn(
                        self.compute_client.spawn_vm,
                        vm_nam, self.image.id,
                        self.flavor,
                        self.network.id,
                        ":" + hosts_lst[i])
                    if deploy_by_hg:
                        threads[(hg_name, hosts_lst[i],
                                 vm_nam)] = child_thread
                    else:
                        threads[(host_to_hg_map.get(hosts_lst[i]),
                                 hosts_lst[i], vm_nam)] = child_thread
                    child_threads.append(child_thread)
                greenthread.sleep(2)
            except eventlet.greenlet.GreenletExit:
                for child_thread in child_threads:
                    child_thread.kill()
            return threads
        except Exception as exc:
            LOG.error(
                f"Error in spawning deploy VM for host group/ hosts: {exc}")
            return threads

    def execute(self):
        try:
            deploy_in_all_hosts_in_pvc = \
                self.conf.deploy_in_all_hosts_in_pvc if hasattr(
                    self.conf, 'deploy_in_all_hosts_in_pvc') else ""
            deploy_on_all_hosts_in_hg = \
                self.conf.deploy_in_all_hosts_in_hg if hasattr(
                    self.conf, 'deploy_in_all_hosts_in_hg') else ""
            build_timeout = self.conf.instance_build_timeout \
                if hasattr(self.conf, 'instance_build_timeout') else 2000
            vm_name = None
            if hasattr(self.conf, "default_prefix"):
                vm_name = self.conf.default_prefix
            if not vm_name:
                vm_name = self.conf.vm_name_prefix
            if (
                deploy_in_all_hosts_in_pvc in ["True", "TRUE", "true"]
                or deploy_on_all_hosts_in_hg in ["True", "TRUE", "true"]
            ):
                threads = {}
                if deploy_in_all_hosts_in_pvc in ["True", "TRUE", "true"]:
                    host_to_hg_map = {}
                    all_hosts_in_pvc = []
                    for agg in self.aggregates:
                        hosts = agg.hosts
                        for host in hosts:
                            all_hosts_in_pvc.append(host)
                            host_to_hg_map[host] = agg.name.lower()
                    threads = self.spawn_multiple_deploys_and_get_threads(
                        all_hosts_in_pvc, host_to_hg_map)
                else:
                    for agg in self.aggregates:
                        if agg.name.lower() == self.host_group.lower():
                            hosts_in_hg = agg.hosts
                            threads = \
                                self.spawn_multiple_deploys_and_get_threads(
                                    hosts_in_hg, hg_name=agg.name.lower())
                            break

                for (hg_name, host_name, vm_nam), thread in six.iteritems(
                        threads):
                    depl_test_res_uuid = uuid.uuid4()
                    try:
                        with eventlet.timeout.Timeout(int(build_timeout)):
                            is_deployed, msg, server = thread.wait()
                            greenthread.sleep(2)
                            if is_deployed:
                                dt_result = self.test_results.DeployTestResult(
                                    test_name=self.name,
                                    status="building",
                                    message="Deploy initiated",
                                    errors={},
                                    vm_name=server.name,
                                    vm_id=server.id,
                                    host_group=hg_name,
                                    host_name=host_name)
                                self.vms.append(server)
                            else:
                                LOG.error(f"Spawning of VM has not"
                                          " returned server object")
                                dt_result = self.test_results.DeployTestResult(
                                    test_name=self.name, status="error",
                                    message="Deploy failed",
                                    errors={"message": msg}, vm_name=vm_nam,
                                    vm_id=None, host_group=hg_name,
                                    host_name=host_name)
                            if server:
                                self.vm_uuid_to_res_uuid_map[server.id] = \
                                    depl_test_res_uuid
                            else:
                                self.vm_uuid_to_res_uuid_map[vm_nam] = \
                                    depl_test_res_uuid
                            self.results[depl_test_res_uuid] = dt_result
                            self.update_depl_result_to_test_flow_result()
                    except (eventlet.timeout.Timeout, Exception) as e:
                        LOG.exception(f"Timed out waiting for deploy VM API"
                                      " to return with error {e}")
                        dt_result = self.test_results.DeployTestResult(
                            test_name=self.name,
                            status="error",
                            message="Deploy failed",
                            errors={
                                "message": msg},
                            vm_name=vm_nam,
                            vm_id=None,
                            host_group=hg_name,
                            host_name=host_name)
                        self.results[depl_test_res_uuid] = dt_result
                        self.vm_uuid_to_res_uuid_map[vm_nam] = \
                            depl_test_res_uuid
                        self.update_depl_result_to_test_flow_result()
            else:
                depl_test_res_uuid = uuid.uuid4()
                vm_name = vm_name + "_" + get_current_datetime()
                if self.host:
                    avaiability_zone = ':' + self.host
                    # TODO: Fetch actual host group if possible
                    hg = self.host_group if self.host_group else None
                    host_name = self.host
                else:
                    avaiability_zone = self.host_group
                    hg = self.host_group
                    host_name = None

                is_deployed, msg, server = self.compute_client.spawn_vm(
                    vm_name, self.image.id, self.flavor,
                    self.network.id, avaiability_zone)
                if is_deployed:
                    self.vms.append(server)
                    self.vm_uuid_to_res_uuid_map[server.id] = \
                        depl_test_res_uuid
                    dt_result = self.test_results.DeployTestResult(
                        test_name=self.name,
                        status="building",
                        message="Deploy initiated",
                        errors={},
                        vm_name=server.name,
                        vm_id=server.id,
                        host_group=hg,
                        host_name=host_name)
                else:
                    LOG.error("Error: Server creation failed.")
                    dt_result = self.test_results.DeployTestResult(
                        test_name=self.name,
                        status="error",
                        message="Deploy failed",
                        errors={
                            "message": msg},
                        vm_name=vm_name,
                        vm_id=None,
                        host_group=hg,
                        host_name=host_name)
                    self.vm_uuid_to_res_uuid_map[vm_name] = depl_test_res_uuid
                self.results[depl_test_res_uuid] = dt_result
                self.update_depl_result_to_test_flow_result()
        except Exception as exc:
            LOG.error(f"Error in execution of deploy API {exc}")

    def wait_for_vm_deploy_to_complete(self, vm):
        try:
            greenthread.sleep(5)
            build_timeout = self.conf.instance_build_timeout \
                if hasattr(self.conf, 'instance_build_timeout') else 2000
            LOG.info(f"Waiting for deploy of VM {vm.name} to complete")
            is_active, err_msg = self.compute_client.wait_for_server_active(
                vm, wait=int(build_timeout))

            if is_active:
                self.update_deploy_results(vm.id, is_active, err_msg)
                self.update_depl_result_to_test_flow_result()
                self.readiness_result.dump_output_file()
                LOG.info(f"VM {vm.id} has become active")
                LOG.info(f"Waiting on VM {vm.id} health to become OK")
                is_ok, msg, h_errors = self.compute_client.pvc_wait_for_vm(
                    vm.id,
                    permissible_statuses=['ACTIVE'],
                    permissible_healths=['OK'],
                    permissible_tasks=[None],
                    interval=10
                )
                self.update_deploy_results(
                    vm.id, is_active, msg, h_errors, health_evaluated=True)
            else:
                self.update_deploy_results(vm.id, is_active, err_msg)
            self.update_depl_result_to_test_flow_result()
            self.readiness_result.dump_output_file()
        except Exception as exc:
            LOG.error(f"Error in waiting for vm deploy "
                      f"to come to active state: {exc}")

    def wait_for_completion(self):
        try:
            pool = GreenPool()
            for vm in self.vms:
                pool.spawn_n(self.wait_for_vm_deploy_to_complete, vm)
            greenthread.sleep(1)
            pool.waitall()
        except Exception as exc:
            LOG.error(f"Error in wait for completion of deploy VM {exc}")

    def update_deploy_results(
            self,
            vm_id,
            is_active,
            err_msg,
            h_errors={},
            health_evaluated=False):
        """
        Updates deploy test result object of a vm with failure, status
        """
        try:
            for res in self.results.values():
                if res.vm_id == vm_id:
                    if not is_active:
                        error_details = \
                            self.compute_client.get_vm_faults_by_vm_id(
                                vm_id)
                        res.status = "failed"
                        res.message = err_msg
                        res.errors = error_details
                    else:
                        if health_evaluated:
                            if h_errors:
                                res.message = err_msg
                                res.errors = h_errors
                                res.status = "deployed"
                            else:
                                res.status = "success"
                                res.message = "Deploy completed successfully"
                        else:
                            res.status = "deployed"
                            res.message = \
                                ("Deploy completed. Waiting for"
                                 " VM health to become OK")
        except Exception as exc:
            LOG.error(f"Error in update_deploy_results exc is {exc}")

    def update_depl_result_to_test_flow_result(self):
        deploy_result = self.get_results()
        self.test_flow_result.add_test_result(self.name, deploy_result)

    def wait_for_delete(self, server, wait=900):
        try:
            clean_up_dict = {
                "status": "success",
                "message": "VM deleted successfully"
            }
            delete_timeout = self.conf.delete_timeout \
                if hasattr(self.conf, 'delete_timeout') else 900
            is_deleted, err_msg = self.compute_client.wait_for_delete(
                server, interval=10, wait=int(delete_timeout))
            res_obj_uuid = self.vm_uuid_to_res_uuid_map[server.id]
            greenthread.sleep(1)
            if is_deleted:
                self.results[res_obj_uuid].clean_up = clean_up_dict
            else:
                vm_delete_faults = self.compute_client.get_vm_faults_by_vm_id(
                    server.id)
                clean_up_dict["status"] = "failed"
                if vm_delete_faults:
                    clean_up_dict["errors"] = vm_delete_faults
                else:
                    clean_up_dict["errors"] = {
                        "message": err_msg
                    }
                self.results[res_obj_uuid].clean_up = clean_up_dict
                raise Exception(err_msg)
        except Exception as exc:
            LOG.error(f"Error in {self.name} wait_for_delete with exc {exc}")
            vm_delete_faults = self.compute_client.get_vm_faults_by_vm_id(
                server.id)
            self.results[res_obj_uuid].clean_up["status"] = "failed"
            self.results[res_obj_uuid].clean_up["errors"] = vm_delete_faults
            raise exc

    def reset_server_status(self, server, state):
        try:
            self.compute_client.reset_server(server, state)
        except Exception as exc:
            LOG.error(
                f"Error in {self.name} reset_server_status with exc {exc}")

    def tear_down(self, server):
        try:
            self.compute_client.delete_server(server)
        except Exception as exc:
            err_msg = (f"Error in delete server for {server.id}"
                       f" Exception is {str(exc)}")
            clean_up_dict = {}
            res_obj_uuid = self.vm_uuid_to_res_uuid_map[server.id]
            vm_delete_faults = self.compute_client.get_vm_faults_by_vm_id(
                server.id)
            clean_up_dict["status"] = "failed"
            if vm_delete_faults:
                clean_up_dict["errors"] = vm_delete_faults
            else:
                clean_up_dict["errors"] = {
                    "message": err_msg
                }
            self.results[res_obj_uuid].clean_up = clean_up_dict
            LOG.error(f"Error in {self.name} teardown with exc {exc}")
            raise exc

    def get_results(self):
        try:
            deploy_vm_result = {}
            deployed_vm_res_lst = []
            len_deploys = len(self.results)
            if len_deploys > 0:
                for result in self.results.values():
                    deploy_vm_result = {
                        "vm_name": result.vm_name,
                        "vm_id": result.vm_id,
                        "status": result.status,
                        "message": result.message,
                        "errors": result.errors,
                        "host_group": result.host_group,
                        "host_name": result.host_name,
                        "clean_up": result.clean_up
                    }
                    deployed_vm_res_lst.append(deploy_vm_result)
            return deployed_vm_res_lst
        except Exception as exc:
            LOG.error(f"Error {exc} in getting deploy results")
            return []

    def get_results_summary(self):
        try:
            error_vm_cnt = 0
            deploy_vm_result = {}
            deployed_vm_res_lst = []
            len_deploys = len(self.results)
            vms_deployed_hosts = []
            failed_host_lst = []
            results_summary = {}
            has_vms_in_warning = False
            warning_msg = (
                " Some VMs which are deployed are in warning state. "
                "Please check RMC status or network configuration.")
            if len_deploys > 0:
                for result in self.results.values():
                    depl_host = result.host_name
                    if result.status == "deployed":
                        has_vms_in_warning = True
                    if (
                        (result.errors and result.status != "deployed")
                            or result.status == 'failed'
                    ):
                        error_vm_cnt += 1
                        if depl_host is not None:
                            failed_host_lst.append(depl_host)
                    if depl_host is not None:
                        vms_deployed_hosts.append(depl_host)
                    deploy_vm_result = {
                        "vm_name": result.vm_name,
                        "vm_id": result.vm_id,
                        "status": result.status,
                        "message": result.message,
                        "errors": result.errors,
                        "host_group": result.host_group,
                        "host_name": result.host_name,
                        "clean_up": result.clean_up
                    }
                    deployed_vm_res_lst.append(deploy_vm_result)
                msg = (str(error_vm_cnt) + " Out of " + str(len_deploys) +
                       " deploys has failed. Hosts where deploy failed"
                       " are " + " ".join(failed_host_lst) + ".")
                if error_vm_cnt > 0:
                    if has_vms_in_warning:
                        final_err_msg = msg + warning_msg
                    else:
                        final_err_msg = msg
                    results_summary["message"] = final_err_msg
                    return False, results_summary, deployed_vm_res_lst
                else:
                    success_msg = ("Deploy of VM is successful on host(s) " +
                                   " ".join(vms_deployed_hosts) + ".")
                    if has_vms_in_warning:
                        final_suc_msg = success_msg + warning_msg
                    else:
                        final_suc_msg = success_msg
                    results_summary["message"] = final_suc_msg
                    return True, results_summary, deployed_vm_res_lst
            else:
                return True, results_summary, deployed_vm_res_lst
        except Exception as exc:
            LOG.error(f"Error {exc} in getting deploy summary results")
            return True, results_summary, deployed_vm_res_lst
