# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

import logging as LOG
import eventlet
# eventlet.monkey_patch()
import uuid
from . import base_test


class MigrateValidateTest(base_test.BaseTest):
    def __init__(self, name, config, blacklist_hosts, compute_client,
                 test_flow_result, readiness_result):
        super(MigrateValidateTest, self).__init__(
            name, config, blacklist_hosts, compute_client=compute_client,
            test_flow_result=test_flow_result,
            readiness_result=readiness_result,
            thread_pool_size=8, api_timeout=180)

    def setUp(self, by_vm=False):
        try:
            self.test_flow_result.update_tests_in_progress(self.name)
            if by_vm:
                vm_name = self.conf.vm_name
                vm_obj = self.compute_client.find_vm(vm_name)
                if not vm_obj:
                    msg = "Could not find the VM"
                    self.update_results(status='skipped',
                                        message=msg, vm_name=vm_name)
                    raise Exception(msg)
                target_host = getattr(self.conf, 'target_host', None)
                if target_host:
                    if not self.compute_client.find_host(target_host):
                        LOG.warn(f"Unable to fetch target host"
                                 f"({target_host}) details from API. It"
                                 " will not be considered for migrate"
                                 " validate target.")
                        # setting it to None
                        target_host = None
                self.execute(vm_obj, target_host)
            else:
                host_group = getattr(self.conf, 'host_group', None)
                if host_group:
                    hg_obj = self.compute_client.find_host_group(host_group)
                    if not hg_obj:
                        msg = (f"The host group '{host_group}' does not"
                               " exist in PowerVC.")
                        LOG.error(msg)
                        return False, msg
                    if len(hg_obj.hosts) > 1:
                        self.setUpByHostGroup(hg_obj)
                    else:
                        hg_name = hg_obj.name
                        err_msg = ("Host group (%s) has one host or no "
                                   "host. Skipping migrate validate on "
                                   "this host group." % hg_name)
                        LOG.warning(err_msg)
                        return False, err_msg
                else:
                    is_retrieved, aggs = self.compute_client.get_aggregates()
                    if not is_retrieved:
                        msg = ("Error: Unable to fetch aggregate "
                               "details from API")
                        LOG.error(msg)
                        return False, msg
                    host_groups = list(aggs)
                    for hg in host_groups:
                        if len(hg.hosts) > 1:
                            self.setUpByHostGroup(hg)
                        else:
                            hg_name = hg.name
                            err_msg = ("Host group (%s) has one host or no "
                                       "host. Skipping migrate validate on "
                                       "this host group." % hg_name)
                            LOG.warning(err_msg)
            return self.isSuccess, "migrate-validate test completed"
        except Exception as exc:
            msg = (f"Error in setUp for MigrateValidateTest: exc is {exc}")
            LOG.error(msg)
            return False, msg

    def setUpByHostGroup(self, host_group):
        hosts = host_group.hosts
        pool = eventlet.GreenPool(self.thread_pool_size)
        for host in hosts:
            if host in self.blacklist_hosts:
                LOG.info("Host (%s) is in blacklist hosts. So skipping "
                         "migrate validate for VMs of this host." % host)
                continue
            vm_list = self.compute_client.list_vms_by_host(host)
            # If VMs are not there on the host, we will check for other hosts
            if not vm_list:
                continue
            target_host_list = [h for h in hosts if h != host]
            for target_host in target_host_list:
                if target_host in self.blacklist_hosts:
                    LOG.info("Host (%s) is in blacklist hosts. So it can "
                             "not be used as target host for "
                             "migrate validate." % target_host)
                    continue
                for vm in vm_list:
                    pool.spawn_n(self.execute, vm, target_host)
                    eventlet.greenthread.sleep(1)
        pool.waitall()

    def execute(self, vm_obj, target_host):
        try:
            vm_id = vm_obj.id
            vm_name = vm_obj.name
            host = vm_obj.hypervisor_hostname
            status = "failed"
            valid_statuses = ['ACTIVE', 'SHUTOFF']
            vm_status = vm_obj.status
            if vm_status not in valid_statuses:
                status = "skipped"
                msg = ("Migrate validate API execution skipped since VM " +
                       vm_id + "is in state " + vm_status)
            else:
                LOG.info(f"Wait for VM {vm_id} status and health "
                         "is OK for migrate validate")
                is_ok, msg, h_errors = self.compute_client.pvc_wait_for_vm(
                    vm_id,
                    permissible_statuses=valid_statuses,
                    permissible_healths=['OK'],
                    permissible_tasks=[None],
                    interval=10
                )
                if is_ok:
                    try:
                        with eventlet.timeout.Timeout(self.api_timeout):
                            isSuccess, result = \
                                self.compute_client.pvc_migrate_validate(
                                    vm_id, target_host)
                            LOG.info("Migrate Validate API response: "
                                     "{%s}." % result)
                            target_host = result.get('target_host',
                                                     target_host)
                            if isSuccess:
                                msg = result.get('message', None)
                                status = "success"
                            else:
                                self.isSuccess = False
                                msg = result.get('error', None)
                                if not msg:
                                    msg = result.get('message', None)
                                LOG.error("Migrate-Validate failed for VM "
                                          "(%s)." % vm_id)
                    except eventlet.timeout.Timeout:
                        self.isSuccess = False
                        msg = ("Migrate Validate API did not complete in "
                               "stipulated time for VM (%s)." % vm_id)
                        LOG.error(msg)
                else:
                    status = "skipped"
                    msg = ("Migrate validate API execution skipped since VM " +
                           vm_id + "health is not OK")
            self.update_results(status=status, message=msg, vm_name=vm_name,
                                vm_id=vm_id, host=host, target_host=target_host)
        except Exception as exc:
            self.isSuccess = False
            LOG.error(f"Error in execution of migrate-validate API {exc}")
    
    def update_results(self, status, message=None, errors=None,
            vm_name=None, vm_id=None, host=None, target_host=None):
        mig_uuid = uuid.uuid4()
        mig_val_result = self.test_results.MigrateValidateResult(
            test_name=self.name, status=status, message=message,
            errors=errors, vm_name=vm_name, vm_id=vm_id, host=host,
            target_host=target_host)
        self.results[mig_uuid] = mig_val_result
        self.update_migv_result_to_test_flow_result()
        self.readiness_result.dump_output_file()

    def get_results(self):
        mig_val_result = {}
        mig_val_result_list = []
        for result in self.results.values():
            mig_val_result = {
                "status": result.status,
                "message": result.message,
                "vm_name": result.vm_name,
                "vm_id": result.vm_id,
                "host": result.host,
                "target_host": result.target_host,
                "errors": result.errors,
            }
            mig_val_result_list.append(mig_val_result)
        return mig_val_result_list

    def update_migv_result_to_test_flow_result(self):
        migrate_result = self.get_results()
        self.test_flow_result.add_test_result(self.name, migrate_result)

    def get_results_summary(self):
        try:
            mig_val_result = {}
            mig_val_result_list = []
            result_summary = {}
            has_failures = False
            failed_migrations_by_src_host = {}
            for result in self.results.values():
                tgt_host = result.target_host
                src_host = result.host
                if result.vm_id:
                    vm_id = result.vm_id
                else:
                    vm_id = result.vm_name

                if result.status == "failed":
                    has_failures = True
                    if src_host not in failed_migrations_by_src_host:
                        failed_migrations_by_src_host[src_host] = {
                            vm_id: [tgt_host]
                        }
                    else:
                        src_host_dict = failed_migrations_by_src_host[src_host]
                        if vm_id in src_host_dict:
                            tgt_host_list = src_host_dict.get(vm_id)
                            tgt_host_list.append(tgt_host)
                        else:
                            src_host_dict[vm_id] = [tgt_host]

                mig_val_result = {
                    "status": result.status,
                    "message": result.message,
                    "vm_name": result.vm_name,
                    "vm_id": result.vm_id,
                    "host": result.host,
                    "target_host": result.target_host,
                    "errors": result.errors,
                }
                mig_val_result_list.append(mig_val_result)
            if has_failures:
                result_summary["message"] = ("Some of Migrate validations"
                                             " has failed. Below provides"
                                             " details at source host level"
                                             " which VMs failed in migrate"
                                             " validate for target host(s)")
                result_summary["details"] = failed_migrations_by_src_host
                return False, result_summary, mig_val_result_list
            else:
                result_summary["message"] = ("All migrate validations"
                                             " were successful")
                return True, result_summary, mig_val_result_list
        except Exception as exc:
            LOG.error(f"Error in migrate validate test summary: {exc}")
            return True, result_summary, mig_val_result_list
