# =================================================================
# Licensed Materials - Property of IBM
#
# (c) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
# =================================================================

from . import deploy_test
from . import attach_volume_test
from . import migrate_validate_test
from powervc_validation_toolkit.results import test_results
import logging as LOG
import eventlet
import six
# eventlet.monkey_patch()
from eventlet import greenthread, GreenPool


class TestFlow:

    def __init__(
            self,
            config,
            compute_client,
            storage_client,
            testflow,
            tests_list):
        """
        TestFlow class gets in config and required
        clients to pass to actual test classes
        """
        self.config = config
        self.compute_client = compute_client
        self.storage_client = storage_client
        self.testflow = testflow
        self.tests_list = tests_list
        self.blacklist_hosts = set()
        # List of Test class objects invoked in test flow
        self.tests_objects = {}
        self.test_summary = test_results.TestSummaryResult()
        self.test_flow_result = test_results.TestFlowResult(
            testflow, self.test_summary)

    def find_deploy_test_object_for_attach(self, attach_test_name):
        index = self.tests_list.index(attach_test_name)
        for i in range(index, 0, -1):
            if self.tests_list[i - 1].startswith('deploy'):
                return self.tests_objects.get(self.tests_list[i - 1])
        return None

    def update_blacklist_hosts(self):
        # We will add hosts to blacklist hosts, that are not in good shape
        # or in maintenance mode
        hosts = self.compute_client.get_hypervisors()
        for host in hosts:
            if host.state != 'up':
                self.blacklist_hosts.add(host.name)
                continue
            if host.status != 'enabled':
                self.blacklist_hosts.add(host.name)

    def setUp(self, readiness_result):
        try:
            # self.update_blacklist_hosts()
            iter_index = 0
            while iter_index < len(self.tests_list):
                curr_test_name = self.tests_list[iter_index]
                LOG.info(f"Started execution of test {curr_test_name}")
                if curr_test_name.startswith('deploy'):
                    # LOG.info(f"Started execution of test {curr_test_name}")
                    depl_test = deploy_test.DeployTest(curr_test_name,
                                                       self.config,
                                                       self.blacklist_hosts,
                                                       self.compute_client,
                                                       self.test_flow_result,
                                                       readiness_result)
                    has_no_errors, msg = depl_test.setUp()
                    if not has_no_errors:
                        self.test_flow_result.update_tests_failed(
                            curr_test_name, msg)
                        LOG.error(f"Test {curr_test_name} is not initiated"
                                  f" with below failure {msg}")
                        depl_failed_result = {
                            "status": "failed",
                            "reason": msg
                        }
                        self.test_flow_result.add_test_result(
                            curr_test_name,
                            depl_failed_result)
                        LOG.info(f"Flow res is "
                                 f"{self.test_flow_result.get_flow_result()}")
                        while (iter_index + 1 < len(self.tests_list) and
                               self.tests_list[iter_index + 1].startswith(
                                   'attach_volume')):
                            attach_msg = ("Deploy has failed with error: " +
                                          msg + " So, attach volume is also "
                                          "not executed.")
                            self.test_flow_result.update_tests_failed(
                                self.tests_list[iter_index + 1], attach_msg)
                            att_result = {
                                "status": "failed", "reason": attach_msg}
                            self.test_flow_result.add_test_result(
                                self.tests_list[iter_index + 1], att_result)
                            iter_index += 1
                        readiness_result.add_flow_result(
                            (self.testflow, self.test_flow_result),
                            self.test_flow_result.get_flow_result())
                        readiness_result.dump_output_file()
                        iter_index += 1
                        continue
                    else:
                        self.test_flow_result.update_tests_in_progress(
                            curr_test_name)
                        depl_test.execute()
                        self.tests_objects[curr_test_name] = depl_test
                        depl_result = \
                            depl_test.get_results()
                        self.test_flow_result.add_test_result(
                            curr_test_name, depl_result)
                        readiness_result.add_flow_result(
                            (self.testflow, self.test_flow_result),
                            self.test_flow_result.get_flow_result())
                        readiness_result.dump_output_file()
                        LOG.info(
                            "Readiness result updated for deploy initiation")
                        no_attach_test_next = True
                        if iter_index + 1 < len(self.tests_list):
                            next_test = self.tests_list[iter_index + 1]
                            if next_test.startswith('attach_volume'):
                                no_attach_test_next = False
                        if len(self.tests_list) == 1 or no_attach_test_next:
                            self.test_flow_result.update_tests_in_progress(
                                curr_test_name)
                            depl_test.wait_for_completion()
                            greenthread.sleep(1)
                            # self.test_flow_result.update_tests_completed(
                            #     curr_test_name, "completed")
                            # readiness_result.add_flow_result(
                            #     (self.testflow, self.test_flow_result),
                            #     self.test_flow_result.get_flow_result())
                            # readiness_result.dump_output_file()
                        iter_index += 1

                elif curr_test_name.startswith('attach_volume'):
                    # LOG.info(f"Started execution of test {curr_test_name}")
                    related_deploy_test_object = \
                        self.find_deploy_test_object_for_attach(curr_test_name)
                    if related_deploy_test_object is not None:
                        # Actual execution required
                        attach_test = attach_volume_test.AttachVolumeTest(
                            curr_test_name,
                            self.config,
                            self.blacklist_hosts,
                            self.compute_client,
                            self.storage_client,
                            related_deploy_test_object,
                            self.test_flow_result,
                            readiness_result)
                        attach_test.setUp()
                        self.tests_objects[curr_test_name] = attach_test
                    else:
                        LOG.error(f"Error in finding matching deploy test"
                                  " for attach volume {curr_test_name}")
                    iter_index += 1

                elif curr_test_name.startswith('migrate_validate'):
                    # LOG.info(f"Started execution of test {curr_test_name}")
                    mig_val_test = migrate_validate_test.MigrateValidateTest(
                        curr_test_name, self.config, self.blacklist_hosts,
                        self.compute_client, self.test_flow_result,
                        readiness_result)
                    if curr_test_name.startswith('migrate_validate_by_vm'):
                        has_no_errors, msg = mig_val_test.setUp(by_vm=True)
                    else:
                        has_no_errors, msg = mig_val_test.setUp()
                    self.tests_objects[curr_test_name] = mig_val_test
                    iter_index += 1

                else:
                    LOG.error(f"{curr_test_name} is not implemented")
                    iter_index += 1

                # iter_index += 1
        except Exception as exc:
            LOG.error(f"TestFlow {self.testflow} execution has failed"
                      f" with exception {exc}")

    def update_flow_results(self, readiness_result):
        try:
            for tes in self.tests_list:
                test_object = self.tests_objects.get(tes)
                if test_object is not None:
                    result = []
                    has_no_errors, res_summary, result = \
                        test_object.get_results_summary()
                    self.test_flow_result.add_test_result(tes, result)
                    if has_no_errors:
                        self.test_flow_result.update_tests_completed(
                            tes, res_summary)
                    else:
                        self.test_flow_result.update_tests_failed(
                            tes, res_summary)

            readiness_result.add_flow_result(
                (self.testflow, self.test_flow_result),
                self.test_flow_result.get_flow_result())
            readiness_result.dump_output_file()
        except Exception as exc:
            LOG.error(f"update_flow_results has failed with exc {exc}")

    def reset_vms_with_exception(self, remaining_vm_ids, test_obj, state):
        try:
            pool = GreenPool(size=len(test_obj.vms))
            threads_list = {}
            for server in test_obj.vms:
                if remaining_vm_ids:
                    if server['id'] not in remaining_vm_ids:
                        LOG.info(f"Skipping VM: {server['id']} for reset")
                        continue
                    else:
                        LOG.info(f"Including VM: {server['id']} for reset")
                threads_list[server['id']] = pool.spawn(
                    test_obj.reset_server_status, server, state)
            thread_responses = {}
            for server_id, thread in six.iteritems(threads_list):
                try:
                    reset_timeout = self.config.delete_timeout \
                        if hasattr(self.config, 'delete_timeout') else 900
                    with eventlet.timeout.Timeout(int(reset_timeout)):
                        LOG.info(("Invoking reset VM REST API for"
                                  " server_id %(server_id)s") % dict(
                            server_id=server_id))
                        thread_responses[server_id] = thread.wait()
                        greenthread.sleep(2)
                except eventlet.timeout.Timeout:
                    msg = (
                        f"Timed out waiting for reset server rest api for VM"
                        f": {server_id}. Killing thread.")
                    LOG.exception(msg)
                    thread.kill()
                    thread_responses[server_id] = 'ResetVMEventletTimeout'
            LOG.info(
                f"reset_vms_with_exception: thread_responses: {thread_responses}")
        except Exception as exc:
            LOG.error(f"Exception occurred in reset "
                      f" state of VMs :{exc}")

    def trigger_delete_for_vms(
            self,
            test_obj,
            reset=False,
            wait_timedout_vms=None):
        w_server_ids = []
        try:
            if reset:
                LOG.info(
                    f"Reseting VMs that are timed out to re-try delete operation")
                if wait_timedout_vms:
                    w_server_ids = [w['id'] for w in wait_timedout_vms]
                    self.reset_vms_with_exception(
                        w_server_ids, test_obj, "active")
                    LOG.info(f"Reset for {w_server_ids} is now "
                             f"complete. Attempting to delete those VMs")
            pool = GreenPool(size=len(test_obj.vms))
            threads_list = {}
            for server in test_obj.vms:
                if w_server_ids:
                    if server['id'] not in w_server_ids:
                        continue
                    else:
                        LOG.info(f"[After-reset-case]Triggering "
                                 f"delete for {server['id']}")
                threads_list[server['id']] = pool.spawn(
                    test_obj.tear_down, server)
            thread_responses = {}
            delete_rest_api_failed_vms = []
            delete_rest_api_timed_out_vms = []
            for server_id, thread in six.iteritems(threads_list):
                try:
                    delete_timeout = self.config.delete_timeout \
                        if hasattr(self.config, 'delete_timeout') else 900
                    with eventlet.timeout.Timeout(int(delete_timeout)):
                        LOG.info(
                            ("Invoking delete REST API for server_id %(server_id)s") %
                            dict(
                                server_id=server_id))
                        thread_responses[server_id] = thread.wait()
                        greenthread.sleep(2)
                except eventlet.timeout.Timeout:
                    msg = (
                        f"Timed out waiting for delete server rest api for VM"
                        f": {server_id}. Killing thread.")
                    LOG.exception(msg)
                    thread.kill()
                    thread_responses[server_id] = 'RestAPITimeout'
                    delete_rest_api_timed_out_vms.append(server_id)

                except Exception as exc:
                    LOG.error(f"Exception found in tear_down : {exc}")
                    delete_rest_api_failed_vms.append(server_id)

            if delete_rest_api_timed_out_vms:
                LOG.info(
                    f"Delete call to VMs timedout: {','.join(delete_rest_api_timed_out_vms)}")
            if delete_rest_api_failed_vms:
                LOG.info(
                    f"Delete call to VMs failed: {','.join(delete_rest_api_failed_vms)}")
            return delete_rest_api_timed_out_vms + delete_rest_api_failed_vms
        except Exception as exc:
            LOG.info(
                f"Exception occurred while triggering delete VMs {exc}")
            return []

    def wait_for_delete_of_vms(self, test_obj, delete_failed_vms):
        wait_timedout_vms = []
        try:
            pool = GreenPool(size=len(test_obj.vms))
            threads_list = {}
            server_objects = {}
            for server in test_obj.vms:
                if server['id'] in delete_failed_vms:
                    continue
                threads_list[server['id']] = pool.spawn(
                    test_obj.wait_for_delete, server)
                server_objects[server['id']] = server
            thread_responses = {}
            for server_id, thread in six.iteritems(threads_list):
                try:
                    delete_timeout = self.config.delete_timeout \
                        if hasattr(self.config, 'delete_timeout') else 900
                    with eventlet.timeout.Timeout(int(delete_timeout)):
                        thread_responses[server_id] = thread.wait()
                        greenthread.sleep(2)
                except eventlet.timeout.Timeout:
                    msg = (
                        f"Timed out waiting for delete server to complete for VM"
                        f": {server_id}. Killing thread.")
                    LOG.exception(msg)
                    thread.kill()
                    thread_responses[server_id] = 'DeleteServerTimeout'
                    wait_timedout_vms.append(server_objects.get(server_id))
                except Exception as exc:
                    LOG.error(
                        f"Exception in wait for deletion of vm {server_id}")
                    wait_timedout_vms.append(server_objects.get(server_id))
            if wait_timedout_vms:
                timedout_vm_list = [vm['id'] for vm in wait_timedout_vms]
                LOG.info(
                    f"Timed out after triggering delete of VMs: {','.join(timedout_vm_list)}")
        except Exception as exc:
            LOG.info(
                f"Exception occurred while waiting for delete of VMs {exc}")
        return wait_timedout_vms

    def tear_down(self):
        try:
            has_deploy_test = False
            for key in self.tests_objects:
                if key.startswith("deploy"):
                    has_deploy_test = True
                    test_obj = self.tests_objects[key]
                    delete_failed_vms = self.trigger_delete_for_vms(test_obj)
                    wait_timedout_vms = self.wait_for_delete_of_vms(
                        test_obj, delete_failed_vms)
                    if wait_timedout_vms:
                        self.trigger_delete_for_vms(
                            test_obj, reset=True,
                            wait_timedout_vms=wait_timedout_vms)
            if not has_deploy_test:
                LOG.info(f"The test flow {self.testflow} does"
                         f" not need any tear down / clean up")
        except Exception as exc:
            LOG.error(f"Error {exc} in tear_down of test flow {self.testflow}")
