#!/bin/bash
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with
# IBM Corp.

# check to see if the shell is bash
# Some kinds of shells choke on compound commands after the || so
# Run once to generate error message and a second time to exit with error code
set +x
/usr/bin/ls -l /proc/$$/exe|/usr/bin/grep bash>/dev/null || echo "Install must be executed using /usr/bin/bash"
/usr/bin/ls -l /proc/$$/exe|/usr/bin/grep bash>/dev/null || exit 42

# Pre-defined Variables
PRODUCT_NAME=powervc_readiness
BIN_DIR=/usr/bin
SBIN_DIR=/usr/sbin
PIP3="${BIN_DIR}/pip3.9"
SITE_PACKAGES_DIR=/usr/lib/python3.9/site-packages
RC=0
red=`tput setaf 1`
reset=`tput sgr0`
OPERATING_SYSTEM=$(cat /etc/*-release | grep -w NAME | cut -d= -f2 | tr -d '"')

usage() {
    echo "Usage: $0 [ -h ]"
    echo ""
    echo "optional arguments:"
    echo "-h, --help            show this help message and exit"
}

INVALID_ARG=0
if [[ $# -gt 0 ]]; then
    if [[ $# -eq 1 ]]; then
        case "$1" in
            --help | -h )
                usage; exit 0 ;;
            * )
	        INVALID_ARG=1 ;;
        esac
    else
        INVALID_ARG=1
    fi
fi

if [[ ${INVALID_ARG} -eq 1 ]]; then
    echo "$0: invalid argument(s) -- $@" 1>&2
    usage; exit 1
fi

# Methods
msg_echo()
{
   ${BIN_DIR}/echo "$*"
}

msginfo_log()
{
   msg_echo "`date +%Y-%m-%dT%H:%M:%S.%Nnano%:z | sed 's/.......nano//g'` [INFO]: $*"
}

msgerr_log()
{
   msg_echo "$red`date +%Y-%m-%dT%H:%M:%S.%Nnano%:z | sed 's/.......nano//g'` [ERROR]: $* $reset"
   msg_end
}

msg_start()
{
    local time=$(/usr/bin/date +%Y-%m-%dT%H:%M:%S%:z)
    msg_echo "################################################################################"
    msg_echo "Starting the ${PRODUCT_NAME} and its dependencies installation on:"
    msg_echo "${time}"
    msg_echo "################################################################################"
    msg_echo " "
}

msg_end()
{
    local time=$(/usr/bin/date +%Y-%m-%dT%H:%M:%S%:z)
    msg_echo " "
    msg_echo "################################################################################"
    if [[ "$RC" -eq 0 ]]; then
      	msg_echo "${PRODUCT_NAME} installation completed at $time."
      	config_file_path="${SITE_PACKAGES_DIR}/powervc_validation_toolkit/conf/powervc_readiness.conf"
      	msg_echo "The config file path is: $config_file_path"
      	msg_echo "You can alter the configuration as per needs and powervc_readiness executable can be run."
	msg_echo "You can also provide your custom configuration file from any directory with -c / --conf option."
    else
      	msg_echo "$red${PRODUCT_NAME} installation failed at $time.$reset"
    fi
    msg_echo "################################################################################"
    exit $RC
}

install_packages()
{
    if [[ "$OPERATING_SYSTEM" == *"Red"* ]]; then
	# Installing python3.9
        msginfo_log "Installing python3.9 ..."
	sudo /usr/bin/yum install python39 -y >/dev/null 2>&1
	RC=$?
	if [[ ${RC} -eq 0 ]]; then
            msginfo_log "python3.9 installed successfully."
        else
            msgerr_log "python3.9 installation failed. Please install it manually before proceeding to ${PRODUCT_NAME} installation."
        fi
	
        # Installing Development Tools
	msginfo_log "Installing Development Tools ..."
        sudo /usr/bin/yum groupinstall 'Development Tools' -y >/dev/null 2>&1
        RC=$?
        if [[ ${RC} -eq 0 ]]; then
	    msginfo_log "Development Tools installed successfully."
	else
	    msgerr_log "Development Tools installation failed. Please install Development Tools via groupinstall before proceeding to ${PRODUCT_NAME} installation."
	fi
    
	# Installing dependenices rust and openssl
	msginfo_log "Installing dependenices rust and openssl ..."
	err_msg="Somthing went wrong, while installing dependenices rust and openssl."
	curl --proto '=https' --tlsv1.2 -sSf https://sh.rustup.rs > /tmp/rustup.sh
	RC=$?
	if [[ ${RC} -ne 0 ]]; then
	    msgerr_log ${err_msg}
	fi
	sudo chmod +x /tmp/rustup.sh
	RC=$?
        if [[ ${RC} -ne 0 ]]; then
            msgerr_log ${err_msg}
        fi
    	sudo /tmp/rustup.sh -y >/dev/null 2>&1
	RC=$?
        if [[ ${RC} -ne 0 ]]; then
            msgerr_log ${err_msg}
        fi
	source $HOME/.cargo/env
	RC=$?
        if [[ ${RC} -ne 0 ]]; then
            msgerr_log ${err_msg}
        fi
    	sudo /usr/bin/yum install openssl-devel -y >/dev/null 2>&1
	RC=$?
        if [[ ${RC} -ne 0 ]]; then
            msgerr_log ${err_msg}
        fi
    	sudo /usr/bin/yum install python39-devel -y >/dev/null 2>&1
	RC=$?
        if [[ ${RC} -ne 0 ]]; then
            msgerr_log ${err_msg}
        fi
	msginfo_log "rust and openssl installed successfully."

	# Installing openstacksdk
        msginfo_log "Installing openstacksdk ..."
	${PIP3} install openstacksdk=="2.1.0" >/dev/null 2>&1
        RC=$?
        if [[ ${RC} -eq 0 ]]; then
            msginfo_log "openstacksdk installed successfully."
        else
            msgerr_log "openstacksdk installation failed."
        fi

	# Installing python nova client
	msginfo_log "Installing python nova client ..."
	sudo ${PIP3} install python-novaclient=="17.4.0" >/dev/null 2>&1
	RC=$?
        if [[ ${RC} -eq 0 ]]; then
            msginfo_log "python nova client installed successfully."
        else
            msgerr_log "python nova client installation failed."
        fi

	# Installing six
	msginfo_log "Installing six ..."
	sudo ${PIP3} install six >/dev/null 2>&1
	RC=$?
        if [[ ${RC} -eq 0 ]]; then
            msginfo_log "six installed successfully."
        else
            msgerr_log "six installation failed."
        fi

	# Installing eventlet
	msginfo_log "Installing eventlet ..."
	sudo ${PIP3} install eventlet >/dev/null 2>&1
	RC=$?
        if [[ ${RC} -eq 0 ]]; then
            msginfo_log "eventlet installed successfully."
        else
            msgerr_log "eventlet installation failed."
        fi
	msginfo_log "Installed all the required dependencies successfully."
    else
	RC=1; msgerr_log "Operating System is not supported."
    fi
}

copy_package()
{
    msginfo_log "Copying powervc_validation_toolkit to site-packages ..."
    folder_name="powervc_validation_toolkit"
    sudo cp -r "$folder_name" "${SITE_PACKAGES_DIR}"
    RC=$?
    if [[ ${RC} -eq 0 ]]; then
        msginfo_log "Successfully copied powervc_validation_toolkit to site-packages."
    else
        msgerr_log "Somthing went wrong, while copying powervc_validation_toolkit to site-packages."
    fi
}

create_binary()
{
msginfo_log "Creating ${PRODUCT_NAME} binary ..."
binary="${BIN_DIR}/${PRODUCT_NAME}"
cat <<EOF >${binary}
#! /usr/bin/python3.9
import sys
from powervc_validation_toolkit.powervc_readiness import main
import pkg_resources
if __name__ == "__main__":
    sys.exit(main())
EOF
RC=$?
if [[ ${RC} -eq 0 ]]; then
    msginfo_log "${PRODUCT_NAME} binary created successfully."
else
    msgerr_log "${PRODUCT_NAME} binary creation failed."
fi
msginfo_log "Setting execute permissions for ${binary} ..."
sudo chmod +x ${binary}
RC=$?
if [[ ${RC} -eq 0 ]]; then
    msginfo_log "Successfully set execute permissions for ${binary}."
else
    msgerr_log "Somthing went wrong, while setting execute permissions for ${binary}."
fi
}
msg_start
install_packages
copy_package
create_binary
msg_end

