#!/bin/bash
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2024 All Rights Reserved
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with
# IBM Corp.

# check to see if the shell is bash
# Some kinds of shells choke on compound commands after the || so
# Run once to generate error message and a second time to exit with error code
set +x
/usr/bin/ls -l /proc/$$/exe|/usr/bin/grep bash>/dev/null || echo "Install must be executed using /usr/bin/bash"
/usr/bin/ls -l /proc/$$/exe|/usr/bin/grep bash>/dev/null || exit 42

# Pre-defined Variables
PRODUCT_NAME=powervc_readiness
BIN_DIR=/usr/bin
SITE_PACKAGES_DIR=/usr/lib/python3.9/site-packages
RC=0
red=`tput setaf 1`
reset=`tput sgr0`
# set default force value to 0, which means false. 1 means true
force=0 # 0 means false

usage() {
    echo "Usage: $0 [ -h ] [ -f ]"
    echo ""
    echo "optional arguments:"
    echo "-f, --force           This will skip the version & build check"
    echo "-h, --help            show this help message and exit"
}

INVALID_ARG=0
if [[ $# -gt 0 ]]; then
    if [[ $# -eq 1 ]]; then
        case "$1" in
            --help | -h )
                usage; exit 0 ;;
	    --force | -f )
		force=1 ;;
            * )
	        INVALID_ARG=1 ;;
        esac
    else
        INVALID_ARG=1
    fi
fi

if [[ ${INVALID_ARG} -eq 1 ]]; then
    echo "$0: invalid argument(s) -- $@" 1>&2
    usage; exit 1
fi

# Methods
msg_echo()
{
   ${BIN_DIR}/echo "$*"
}

msginfo_log()
{
   msg_echo "`date +%Y-%m-%dT%H:%M:%S.%Nnano%:z | sed 's/.......nano//g'` [INFO]: $*"
}

msgerr_log()
{
   msg_echo "$red`date +%Y-%m-%dT%H:%M:%S.%Nnano%:z | sed 's/.......nano//g'` [ERROR]: $* $reset"
   msg_end
}

msgwarn_log()
{
   msg_echo "`date +%Y-%m-%dT%H:%M:%S.%Nnano%:z | sed 's/.......nano//g'` [WARNING]: $*"
   while true; do
      read -p "Would you like to proceed?" yn
      case $yn in
        [Yy]* ) break;;
        [Nn]* ) exit;;
        * ) echo "Please answer yes or no.";;
      esac
   done
}

msg_start()
{
    local time=$(/usr/bin/date +%Y-%m-%dT%H:%M:%S%:z)
    msg_echo "################################################################################"
    msg_echo "Starting the ${PRODUCT_NAME} update on:"
    msg_echo "${time}"
    msg_echo "################################################################################"
    msg_echo " "
}

msg_end()
{
    local time=$(/usr/bin/date +%Y-%m-%dT%H:%M:%S%:z)
    msg_echo " "
    msg_echo "################################################################################"
    if [[ "$RC" -eq 0 ]]; then
      	msg_echo "${PRODUCT_NAME} update completed at $time."
      	config_file_path="${SITE_PACKAGES_DIR}/powervc_validation_toolkit/conf/powervc_readiness.conf"
      	msg_echo "The config file path is: $config_file_path"
      	msg_echo "You can alter the configuration as per needs and powervc_readiness executable can be run."
	msg_echo "You can also provide your custom configuration file from any directory with -c / --conf option."
    else
      	msg_echo "$red${PRODUCT_NAME} update failed at $time.$reset"
    fi
    msg_echo "################################################################################"
    exit $RC
}

copy_package()
{
    msginfo_log "Copying powervc_validation_toolkit to site-packages ..."
    folder_name="powervc_validation_toolkit"
    sudo cp -r "$folder_name" "${SITE_PACKAGES_DIR}"
    RC=$?
    if [[ ${RC} -eq 0 ]]; then
        msginfo_log "Successfully copied powervc_validation_toolkit to site-packages."
    else
        msgerr_log "Somthing went wrong, while copying powervc_validation_toolkit to site-packages."
    fi
}

compare_versions() {
  ver1="$1"
  ver2="$2"

  # same version return 2
  if [ "$ver1" == "$ver2" ]; then
    return 2
  fi

  IFS='.' read -r -a v1 <<< "$ver1"
  IFS='.' read -r -a v2 <<< "$ver2"

  for i in "${!v1[@]}"; do
    if [ "${v1[$i]}" -gt "${v2[$i]}" ]; then
      # version 1 is greater than version 2 return 0
      return 0
    elif [ "${v1[$i]}" -lt "${v2[$i]}" ]; then
      # version 1 is lesser than version 2 return 1
      return 1
    fi
  done

  return 2
}

compare_build() {
    build1="$1"
    build2="$2"

    # same build return 2
    if [ "$build1" == "$build2" ]; then
        return 2
    fi

    # Converting to timestamp/date format
    arr=(${build1//-/ })
    t=`echo ${arr[1]}00|sed 's/../&:/g; s/:$//'`
    timestamp1=${arr[0]}" "$t

    arr=(${build2//-/ })
    t=`echo ${arr[1]}00|sed 's/../&:/g; s/:$//'`
    timestamp2=${arr[0]}" "$t

    # Converting to seconds
    t1=$(date --date="$timestamp1" +%s)
    t2=$(date --date="$timestamp2" +%s)

    if [ "$t1" -gt "$t2" ]; then
      # build 1 is greater than build 2 return 0
      return 0
    else
      # build 1 is lesser than build 2 return 1
      return 1
    fi
}


if [ $force -eq 1 ]; then
    msgwarn_log "force option will skip the version & build check, which is not recommended."
else
    # fetch current build
    current_version_build=`/usr/bin/powervc_readiness -V`
    RC=$?
    if [[ ${RC} -ne 0 ]]; then
        msgerr_log "Unable to fetch the current powervc_readiness version and build details."
    fi
    arr=(${current_version_build// / })
    # fetch new build
    new_version=`cat powervc_validation_toolkit/version.properties |grep version|tail -1|cut -d= -f2|tr -d " "`
    RC=$?
    if [[ ${RC} -ne 0 ]]; then
        msgerr_log "Somthing went wrong while fetching the new powervc_readiness version."
    fi
    new_build=`cat powervc_validation_toolkit/version.properties |grep build|tail -1|cut -d= -f2|tr -d " "`
    RC=$?
    if [[ ${RC} -ne 0 ]]; then
        msgerr_log "Somthing went wrong while fetching the new powervc_readiness build."
    fi

    # compare versions
    compare_versions $new_version ${arr[0]}
    RC=$?
    if [[ ${RC} -eq 1 ]]; then
        msgerr_log "Current powervc_readiness version should not be greater than the new version."
    elif [[ ${RC} -eq 2 ]]; then
	# vsersion is same, compare builds
	compare_build $new_build ${arr[1]}
	RC=$?
	if [[ ${RC} -ne 0 ]]; then
            msgerr_log "New powervc_readiness build should not be greater than the current build."
	fi
    fi
fi
msg_start
copy_package
msg_end
